#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Language modeling trainer unit tests."""

from typing import cast

import sentencepiece as _sentencepiece
import importlib_resources
from pytorch_lightning.callbacks.model_checkpoint import ModelCheckpoint

from gt4sd.training_pipelines import (
    TRAINING_PIPELINE_MAPPING,
    LanguageModelingTrainingPipeline,
)
from gt4sd.training_pipelines.pytorch_lightning.language_modeling.lm_datasets import (
    CGMDataModule,
    CLMDataModule,
    MLMDataModule,
    PLMDataModule,
)
from gt4sd.training_pipelines.pytorch_lightning.language_modeling.models import (
    CGMModule,
    CLMModule,
    MLMModule,
    PLMModule,
)

# sentencepiece has to be loaded before lightning to avoid segfaults
_sentencepiece

template_config = {
    "model_args": {
        "tokenizer": "albert-base-v2",
        "model_name_or_path": "albert-base-v2",
        "model_config_name": "albert-base-v2",
        "type": "mlm",
        "lr": 2e-5,
        "lr_decay": 0.5,
        "cache_dir": "/tmp/dci_",
    },
    "dataset_args": {
        "max_length": 512,
        "mlm_probability": 0.15,
        "plm_probability": 0.1666,
        "max_span_length": 5,
        "batch_size": 8,
        "train_file": "ade_corpus_v2",
        "validation_file": "ade_corpus_v2",
        "cache_dir": "/tmp/dci_",
    },
    "trainer_args": {
        "default_root_dir": "here",
        "val_check_interval": 5000,
        "max_epochs": 1,
        "accumulate_grad_batches": 1,
        "limit_val_batches": 500,
        "log_every_n_steps": 500,
        "monitor": "val_loss",
        "save_top_k": 2,
        "mode": "min",
        "every_n_train_steps": 50000,
    },
}


def test_get_data_and_model_modules_mlm():

    pipeline = TRAINING_PIPELINE_MAPPING.get("language-modeling-trainer")

    assert pipeline is not None

    test_pipeline = cast(LanguageModelingTrainingPipeline, pipeline())

    config = template_config.copy()

    with importlib_resources.as_file(
        importlib_resources.files("gt4sd") / "training_pipelines/tests/lm_example.jsonl"
    ) as file_path:

        config["dataset_args"]["train_file"] = file_path

        config["dataset_args"]["validation_file"] = file_path
        config["model_args"]["type"] = "mlm"

        data_module, model_module = test_pipeline.get_data_and_model_modules(
            config["model_args"], config["dataset_args"]  # type: ignore
        )

        assert isinstance(model_module, MLMModule)
        assert isinstance(data_module, MLMDataModule)

        check_model_config(model_module, config["model_args"])
        check_data_config(data_module, config["dataset_args"])


def test_get_data_and_model_modules_clm():

    pipeline = TRAINING_PIPELINE_MAPPING.get("language-modeling-trainer")

    assert pipeline is not None

    test_pipeline = cast(LanguageModelingTrainingPipeline, pipeline())

    config = template_config.copy()

    with importlib_resources.as_file(
        importlib_resources.files("gt4sd") / "training_pipelines/tests/lm_example.jsonl"
    ) as file_path:

        config["dataset_args"]["train_file"] = file_path
        config["dataset_args"]["validation_file"] = file_path
        config["model_args"]["type"] = "clm"
        config["model_args"]["model_name_or_path"] = "gpt2"

        data_module, model_module = test_pipeline.get_data_and_model_modules(
            config["model_args"], config["dataset_args"]  # type: ignore
        )

        assert isinstance(model_module, CLMModule)
        assert isinstance(data_module, CLMDataModule)

        check_model_config(model_module, config["model_args"])
        check_data_config(data_module, config["dataset_args"])


def test_get_data_and_model_modules_cgm():

    pipeline = TRAINING_PIPELINE_MAPPING.get("language-modeling-trainer")

    assert pipeline is not None

    test_pipeline = cast(LanguageModelingTrainingPipeline, pipeline())

    config = template_config.copy()

    with importlib_resources.as_file(
        importlib_resources.files("gt4sd") / "training_pipelines/tests/lm_example.jsonl"
    ) as file_path:

        config["dataset_args"]["train_file"] = file_path
        config["dataset_args"]["validation_file"] = file_path
        config["model_args"]["type"] = "cgm"
        config["model_args"]["model_name_or_path"] = "t5-base"

        data_module, model_module = test_pipeline.get_data_and_model_modules(
            config["model_args"], config["dataset_args"]  # type: ignore
        )

        assert isinstance(model_module, CGMModule)
        assert isinstance(data_module, CGMDataModule)

        check_model_config(model_module, config["model_args"])
        check_data_config(data_module, config["dataset_args"])


def test_get_data_and_model_modules_plm():

    pipeline = TRAINING_PIPELINE_MAPPING.get("language-modeling-trainer")

    assert pipeline is not None

    test_pipeline = cast(LanguageModelingTrainingPipeline, pipeline())

    config = template_config.copy()

    with importlib_resources.as_file(
        importlib_resources.files("gt4sd") / "training_pipelines/tests/lm_example.jsonl"
    ) as file_path:

        config["dataset_args"]["train_file"] = file_path
        config["dataset_args"]["validation_file"] = file_path
        config["model_args"]["type"] = "plm"
        config["model_args"]["model_name_or_path"] = "xlnet-base-cased"

        data_module, model_module = test_pipeline.get_data_and_model_modules(
            config["model_args"], config["dataset_args"]  # type: ignore
        )

        assert isinstance(model_module, PLMModule)
        assert isinstance(data_module, PLMDataModule)

        check_model_config(model_module, config["model_args"])
        check_data_config(data_module, config["dataset_args"])


def test_add_callbacks():

    pipeline = TRAINING_PIPELINE_MAPPING.get("language-modeling-trainer")

    assert pipeline is not None

    callbacks_input = {
        "model_checkpoint_callback": {
            "monitor": "val_loss",
            "save_top_k": 2,
            "mode": "min",
            "every_n_train_steps": 50000,
        }
    }

    callbacks = pipeline().add_callbacks(callbacks_input)  # type: ignore

    assert len(callbacks) == 1
    assert isinstance(callbacks[0], ModelCheckpoint)


def check_model_config(module, config):
    for entry in module.model_args:
        assert entry in config
        assert config[entry] == module.model_args[entry]


def check_data_config(module, config):
    for entry in module.dataset_args:
        assert entry in config
        assert config[entry] == module.dataset_args[entry]



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Moses AAE implementation."""

import argparse

from guacamol_baselines.moses_baselines.aae_distribution_learning import AaeGenerator


class AAE:
    def __init__(
        self,
        model_path: str,
        model_config_path: str,
        vocab_path: str,
        n_samples: int,
        n_batch: int,
        max_len: int,
        device: str = "cpu",
    ):
        """Initialize AAE.

        Args:
            model_path: path from where to load the model.
            model_config_path: path from where to load the model config.
            vocab_path: path from where to load the vocab.
            n_samples: number of samples to sample.
            n_batch: size of the batch.
            max_len: max length of SMILES.
            device: device used for computation. Defaults to cpu.
        """
        self.config = argparse.Namespace(
            model_load=model_path,
            config_load=model_config_path,
            vocab_load=vocab_path,
            n_samples=n_samples,
            n_batch=n_batch,
            max_len=max_len,
            device=device,
        )

    def get_generator(self) -> AaeGenerator:
        """Create an instance of the AaeGenerator.

        Returns:
            an instance of AaeGenerator.
        """
        optimiser = AaeGenerator(self.config)
        return optimiser



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Data module for granular."""

import logging
from typing import Callable, List, Optional, cast

import sentencepiece as _sentencepiece
import pandas as pd
import pytorch_lightning as pl
import torch
from torch.utils.data import DataLoader, Sampler, Subset, random_split

from .dataset import CombinedGranularDataset, GranularDataset
from .sampler import StratifiedSampler

# sentencepiece has to be loaded before lightning to avoid segfaults
_sentencepiece

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class GranularDataModule(pl.LightningDataModule):
    """Data module from granular."""

    def __init__(
        self,
        dataset_list: List[GranularDataset],
        validation_split: Optional[float] = None,
        validation_indices_file: Optional[str] = None,
        stratified_batch_file: Optional[str] = None,
        stratified_value_name: Optional[str] = None,
        batch_size: int = 64,
        num_workers: int = 1,
    ) -> None:
        """Construct GranularDataModule.

        Args:
            dataset_list: a list of granular datasets.
            validation_split: proportion used for validation. Defaults to None,
                a.k.a., use indices file if provided otherwise uses half of the data for validation.
            validation_indices_file: indices to use for validation. Defaults to None, a.k.a.,
                use validation split proportion, if not provided uses half of the data for validation.
            stratified_batch_file: stratified batch file for sampling. Defaults to None, a.k.a.,
                no stratified sampling.
            stratified_value_name: stratified value name. Defaults to None, a.k.a.,
                no stratified sampling. Needed in case a stratified batch file is provided.
            batch_size: batch size. Defaults to 64.
            num_workers: number of workers. Defaults to 1.
        """
        super().__init__()
        self.batch_size = batch_size
        self.validation_split = validation_split
        self.validation_indices_file = validation_indices_file
        self.dataset_list = dataset_list
        self.num_workers = num_workers
        self.stratified_batch_file = stratified_batch_file
        self.stratified_value_name = stratified_value_name
        self.prepare_train_data()

    @staticmethod
    def combine_datasets(
        dataset_list: List[GranularDataset],
    ) -> CombinedGranularDataset:
        """Combine granular datasets.

        Args:
            dataset_list: a list of granular datasets.

        Returns:
            a combined granular dataset.
        """
        return CombinedGranularDataset(
            [a_dataset.dataset for a_dataset in dataset_list]
        )

    def prepare_train_data(self) -> None:
        """Prepare training dataset."""
        self.train_dataset = GranularDataModule.combine_datasets(self.dataset_list)

    def prepare_test_data(self, dataset_list: List[GranularDataset]) -> None:
        """Prepare testing dataset.

        Args:
            dataset_list: a list of granular datasets.
        """
        self.test_dataset = GranularDataModule.combine_datasets(dataset_list)

    def setup(self, stage: Optional[str] = None) -> None:
        """Setup the data module.

        Args:
            stage: stage considered, unused. Defaults to None.
        """
        if (
            self.stratified_batch_file is not None
            and self.stratified_value_name is None
        ):
            raise ValueError(
                f"stratified_batch_file={self.stratified_batch_file}, need to set stratified_value_name as well"
            )
        if self.validation_indices_file is None and self.validation_split is None:
            self.validation_split = 0.5
        if self.validation_indices_file:
            val_indices = (
                pd.read_csv(self.validation_indices_file).values.flatten().tolist()
            )
            train_indices = [
                i for i in range(len(self.train_dataset)) if i not in val_indices
            ]
            self.train_data = Subset(self.train_dataset, train_indices)
            self.val_data = Subset(self.train_dataset, val_indices)

        else:
            val = int(len(self.train_dataset) * cast(float, (self.validation_split)))
            train = len(self.train_dataset) - val
            self.train_data, self.val_data = random_split(
                self.train_dataset, [train, val]
            )
        logger.info(f"number of data points used for training: {len(self.train_data)}")
        logger.info(f"number of data points used for validation: {len(self.val_data)}")
        logger.info(
            f"validation proportion: {len(self.val_data) / (len(self.val_data) + len(self.train_data))}"
        )

    @staticmethod
    def get_stratified_batch_sampler(
        stratified_batch_file: str,
        stratified_value_name: str,
        batch_size: int,
        selector_fn: Callable[[pd.DataFrame], pd.DataFrame],
    ) -> StratifiedSampler:
        """Get stratified batch sampler.

        Args:
            stratified_batch_file: stratified batch file for sampling.
            stratified_value_name: stratified value name.
            batch_size: batch size.
            selector_fn: selector function for stratified sampling.
        Returns:
            a stratified batch sampler.
        """
        stratified_batch_dataframe = pd.read_csv(stratified_batch_file)
        stratified_data = stratified_batch_dataframe[
            selector_fn(stratified_batch_dataframe)
        ][stratified_value_name].values
        stratified_data_tensor = torch.from_numpy(stratified_data)
        return StratifiedSampler(targets=stratified_data_tensor, batch_size=batch_size)

    def train_dataloader(self) -> DataLoader:
        """Get a training data loader.

        Returns:
            a training data loader.
        """
        sampler: Optional[Sampler] = None
        if self.stratified_batch_file:
            sampler = GranularDataModule.get_stratified_batch_sampler(
                stratified_batch_file=self.stratified_batch_file,
                stratified_value_name=str(self.stratified_value_name),
                batch_size=self.batch_size,
                selector_fn=lambda dataframe: ~dataframe["validation"],
            )
        return DataLoader(
            self.train_data,
            num_workers=self.num_workers,
            batch_size=self.batch_size,
            pin_memory=False,
            sampler=sampler,
        )

    def val_dataloader(self) -> DataLoader:
        """Get a validation data loader.

        Returns:
            a validation data loader.
        """
        sampler: Optional[Sampler] = None
        if self.stratified_batch_file:
            sampler = GranularDataModule.get_stratified_batch_sampler(
                stratified_batch_file=self.stratified_batch_file,
                stratified_value_name=str(self.stratified_value_name),
                batch_size=self.batch_size,
                selector_fn=lambda dataframe: dataframe["validation"],
            )
        return DataLoader(
            self.val_data,
            num_workers=self.num_workers,
            batch_size=self.batch_size,
            pin_memory=False,
            sampler=sampler,
        )

    def test_dataloader(self) -> DataLoader:
        """Get a testing data loader.

        Returns:
            a testing data loader.
        """
        return DataLoader(
            self.test_dataset,
            batch_size=self.batch_size,
            num_workers=self.num_workers,
            pin_memory=False,
        )



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Patent Generative Transformer (PGT) generation algorithm."""

import logging
import os
import shutil
from dataclasses import field
from typing import Any, ClassVar, Dict, Optional, TypeVar

from typing_extensions import Protocol, runtime_checkable

from ....cli.pl_to_hf_converter import convert_pl_to_hf
from ....training_pipelines.core import TrainingPipelineArguments
from ....training_pipelines.pytorch_lightning.language_modeling.core import (
    LanguageModelingSavingArguments,
)
from ...core import AlgorithmConfiguration, GeneratorAlgorithm, Untargeted
from ...registry import ApplicationsRegistry
from .implementation import (
    COHERENCE_TYPES,
    EDITING_TYPES,
    GENERATION_PROMPTS,
    CoherenceCheckGenerator,
    EditGenerator,
    Generator,
    PartGenerator,
)

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

T = type(None)
S = TypeVar("S", bound=str)


class PGT(GeneratorAlgorithm[S, T]):
    """PGT Algorithm."""

    def __init__(
        self,
        configuration: AlgorithmConfiguration[S, T],
        target: Optional[T] = None,
    ) -> None:
        """Instantiate PGT ready to generate items.

        Args:
            configuration: domain and application
                specification defining parameters, types and validations.
            target: unused since it is not a conditional generator.

        Example:
            An example for generating abstract from a given claim:

                config = PGTGenerator(task="claim_to_abstract", input_text="My interesting claim")
                generator = PGT(configuration=config)
                print(list(generator.sample(1)))
        """

        configuration = self.validate_configuration(configuration)

        self.max_samples = configuration.num_return_sequences  # type: ignore

        # No validation/check on the target input here, since model is not yet loaded.
        super().__init__(
            configuration=configuration,  # type:ignore
            target=target,  # type:ignore
        )

    def get_generator(
        self,
        configuration: AlgorithmConfiguration[S, T],
        target: Optional[T],
    ) -> Untargeted:
        """Get the function to sample with the given configuration.

        Args:
            configuration: helps to set up specific application of PGT.
            target: context or condition for the generation. Unused in the algorithm.

        Returns:
            callable with target generating a batch of items.
        """
        logger.info("ensure artifacts for the application are present.")
        self.local_artifacts = configuration.ensure_artifacts()
        implementation: Generator = configuration.get_generator(  # type: ignore
            self.local_artifacts
        )

        return implementation.generate_case  # type: ignore

    def validate_configuration(
        self, configuration: AlgorithmConfiguration[S, T]
    ) -> AlgorithmConfiguration[S, T]:
        @runtime_checkable
        class AnyPGTConfiguration(Protocol):
            """Protocol for PGT configurations."""

            def get_generator(self, resources_path: str) -> Generator:
                ...

            def validate_item(self, item: Any) -> S:
                ...

        # TODO raise InvalidAlgorithmConfiguration
        assert isinstance(configuration, AnyPGTConfiguration)
        assert isinstance(configuration, AlgorithmConfiguration)
        return configuration


@ApplicationsRegistry.register_algorithm_application(PGT)
class PGTAlgorithmConfiguration(AlgorithmConfiguration[str, None]):
    """Basic configuration for a PGT algorithm"""

    algorithm_type: ClassVar[str] = "generation"
    domain: ClassVar[str] = "nlp"
    algorithm_version: str = "v0"

    model_type: str = field(
        default="",
        metadata=dict(description="Type of the model."),
    )
    max_length: int = field(
        default=512, metadata=dict(description="Maximum length of the generated text.")
    )
    top_k: int = field(
        default=50,
        metadata=dict(description="Number of top-k probability tokens to keep."),
    )
    top_p: float = field(
        default=1.0,
        metadata=dict(
            description="Only tokens with cumulative probabilities summing up to this value are kept."
        ),
    )
    num_return_sequences: int = field(
        default=3,
        metadata=dict(description="Number of alternatives to be generated."),
    )
    no_repeat_ngram_size: int = field(
        default=2,
        metadata=dict(description="Size of n-gram to not appear twice."),
    )

    def get_target_description(self) -> Optional[Dict[str, str]]:
        """Get description of the target for generation.

        Returns:
            target description, returns None in case no target is used.
        """
        return None

    def get_generator(self, resources_path: str, **kwargs) -> Generator:
        """Instantiate the actual PGT implementation.

        Args:
               resources_path: local path to model files.

        Returns:
               instance with
                :meth:`generate_batch<gt4sd.algorithms.generation.pgt.implementation.Generator.generate_case>`
                 method for targeted generation.
        """
        return Generator(
            resources_path=resources_path,
            model_type=self.model_type,
            model_name=self.algorithm_version,
            max_length=self.max_length,
            top_k=self.top_k,
            top_p=self.top_p,
            num_return_sequences=self.num_return_sequences,
        )

    @classmethod
    def save_version_from_training_pipeline_arguments_postprocess(
        cls,
        training_pipeline_arguments: TrainingPipelineArguments,
    ):
        """Postprocess after saving. Remove temporarily converted hf model
           if pytorch-lightning checkpoint is given.

        Args:
            training_pipeline_arguments: training pipeline arguments.
        """

        if isinstance(training_pipeline_arguments, LanguageModelingSavingArguments):
            if training_pipeline_arguments.ckpt is not None:
                shutil.rmtree(training_pipeline_arguments.hf_model_path)

                logger.info(
                    f"Cleaning up temporary files from {training_pipeline_arguments.hf_model_path}"
                )
        else:
            return super().save_version_from_training_pipeline_arguments_postprocess(
                training_pipeline_arguments
            )

    @classmethod
    def get_filepath_mappings_for_training_pipeline_arguments(
        cls, training_pipeline_arguments: TrainingPipelineArguments
    ) -> Dict[str, str]:
        """Ger filepath mappings for the given training pipeline arguments.

        Args:
            training_pipeline_arguments: training pipeline arguments.

        Returns:
            a mapping between artifacts' files and training pipeline's output files.
        """

        if isinstance(training_pipeline_arguments, LanguageModelingSavingArguments):

            if training_pipeline_arguments.ckpt is not None:

                convert_pl_to_hf(training_pipeline_arguments)

            model_files = os.listdir(training_pipeline_arguments.hf_model_path)

            model_files_dict = {
                file: os.path.join(training_pipeline_arguments.hf_model_path, file)
                for file in model_files
            }
            return model_files_dict

        else:
            return super().get_filepath_mappings_for_training_pipeline_arguments(
                training_pipeline_arguments
            )


@ApplicationsRegistry.register_algorithm_application(PGT)
class PGTGenerator(PGTAlgorithmConfiguration):
    """Configuration for a PGT Generator algorithm"""

    input_text: str = field(
        default="This is my input",
        metadata=dict(description="Input text."),
    )
    task: str = field(
        default="title-to-abstract",
        metadata=dict(
            description=f"Generation tasks. Supported: {', '.join(GENERATION_PROMPTS.keys())}"
        ),
    )

    def get_generator(self, resources_path: str, **kwargs) -> Generator:
        """Instantiate the actual PGT implementation for part of patent generation.

        Args:
           resources_path: local path to model files.

        Returns:
           instance with
            :meth:`generate_batch<gt4sd.algorithms.generation.pgt.implementation.Generator.generate_case>`
             method for targeted generation.
        """

        return PartGenerator(
            resources_path=resources_path,
            input_text=self.input_text,
            model_type=self.model_type,
            model_name=self.algorithm_version,
            max_length=self.max_length,
            top_k=self.top_k,
            top_p=self.top_p,
            num_return_sequences=self.num_return_sequences,
            no_repeat_ngram_size=self.no_repeat_ngram_size,
            task=self.task,
        )


@ApplicationsRegistry.register_algorithm_application(PGT)
class PGTEditor(PGTAlgorithmConfiguration):
    """Configuration for a PGT Editor algorithm."""

    input_text: str = field(
        default="This is my input",
        metadata=dict(description="Input text."),
    )
    input_type: str = field(
        default="abstract",
        metadata=dict(
            description=f"Part of a patent the input text belongs. Supported: {', '.join(EDITING_TYPES)}"
        ),
    )

    def get_generator(self, resources_path: str, **kwargs) -> Generator:
        """Instantiate the actual PGT implementation for part of patent editing.

        Args:
           resources_path: local path to model files.

        Returns:
           instance with
            :meth:`generate_batch<gt4sd.algorithms.generation.pgt.implementation.Generator.generate_case>`
             method for targeted generation.
        """

        return EditGenerator(
            resources_path=resources_path,
            input_text=self.input_text,
            model_type=self.model_type,
            model_name=self.algorithm_version,
            max_length=self.max_length,
            top_k=self.top_k,
            top_p=self.top_p,
            num_return_sequences=self.num_return_sequences,
            no_repeat_ngram_size=self.no_repeat_ngram_size,
            input_type=self.input_type,
        )


@ApplicationsRegistry.register_algorithm_application(PGT)
class PGTCoherenceChecker(PGTAlgorithmConfiguration):
    """Configuration for a PGT coherence check algorithm"""

    num_return_sequences: int = field(
        default=1,
        metadata=dict(
            description="Number of alternatives should be always 1 for coherence check."
        ),
    )

    input_a: str = field(
        default="I'm a stochastic parrot.",
        metadata=dict(description="First input for coherence check."),
    )

    input_b: str = field(
        default="I'm a stochastic parrot.",
        metadata=dict(description="Second input for coherence check."),
    )

    coherence_type: str = field(
        default="title-abstract",
        metadata=dict(
            description=f"Input types for the check. Supported: {', '.join(COHERENCE_TYPES)}"
        ),
    )

    def get_generator(self, resources_path: str, **kwargs) -> Generator:
        """Instantiate the actual PGT implementation for patent coherence check.

        Args:
           resources_path: local path to model files.

        Returns:
           instance with
            :meth:`generate_batch<gt4sd.algorithms.generation.pgt.implementation.Generator.generate_case>`
            method for targeted generation.
        """

        return CoherenceCheckGenerator(
            resources_path=resources_path,
            input_a=self.input_a,
            input_b=self.input_b,
            model_type=self.model_type,
            model_name=self.algorithm_version,
            max_length=self.max_length,
            top_k=self.top_k,
            top_p=self.top_p,
            num_return_sequences=self.num_return_sequences,
            no_repeat_ngram_size=self.no_repeat_ngram_size,
            coherence_type=self.coherence_type,
        )



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
from typing import Dict, Tuple, Type, Union

from ...algorithms.core import PredictorAlgorithm
from ..core import PropertyPredictor, PropertyPredictorParameters
from .core import (
    AbsoluteEnergy,
    AbsoluteEnergyParameters,
    BandGap,
    BandGapParameters,
    BulkModuli,
    BulkModuliParameters,
    FermiEnergy,
    FermiEnergyParameters,
    FormationEnergy,
    FormationEnergyParameters,
    MetalNonMetalClassifier,
    MetalNonMetalClassifierParameters,
    MetalSemiconductorClassifier,
    MetalSemiconductorClassifierParameters,
    PoissonRatio,
    PoissonRatioParameters,
    ShearModuli,
    ShearModuliParameters,
)

CRYSTALS_PROPERTY_PREDICTOR_FACTORY: Dict[
    str,
    Tuple[
        Union[Type[PropertyPredictor], Type[PredictorAlgorithm]],
        Type[PropertyPredictorParameters],
    ],
] = {
    #  ml predicted properties
    "formation_energy": (FormationEnergy, FormationEnergyParameters),
    "absolute_energy": (AbsoluteEnergy, AbsoluteEnergyParameters),
    "band_gap": (BandGap, BandGapParameters),
    "fermi_energy": (FermiEnergy, FermiEnergyParameters),
    "bulk_moduli": (BulkModuli, BulkModuliParameters),
    "shear_moduli": (ShearModuli, ShearModuliParameters),
    "poisson_ratio": (PoissonRatio, PoissonRatioParameters),
    "metal_semiconductor_classifier": (
        MetalSemiconductorClassifier,
        MetalSemiconductorClassifierParameters,
    ),
    "metal_nonmetal_classifier": (
        MetalNonMetalClassifier,
        MetalNonMetalClassifierParameters,
    ),
}


AVAILABLE_CRYSTALS_PROPERTY_PREDICTOR = sorted(
    CRYSTALS_PROPERTY_PREDICTOR_FACTORY.keys()
)



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Utilities for algorithms retrieval."""

from typing import Dict, List

from ..algorithms.registry import ApplicationsRegistry, ConfigurationTuple
from ..configuration import reset_logging_root_logger

reset_logging_root_logger()

AVAILABLE_ALGORITHMS = sorted(
    ApplicationsRegistry.list_available(),
    key=lambda algorithm: (
        algorithm["domain"],
        algorithm["algorithm_type"],
        algorithm["algorithm_name"],
        algorithm["algorithm_application"],
        algorithm["algorithm_version"],
    ),
)
AVAILABLE_ALGORITHMS_CATEGORIES = {
    category: sorted(set([algorithm[category] for algorithm in AVAILABLE_ALGORITHMS]))
    for category in ["domain", "algorithm_type"]
}


def filter_algorithm_applications(
    algorithms: List[Dict[str, str]], filters: Dict[str, str]
) -> List[Dict[str, str]]:
    """
    Returning algorithms with given filters.

    Args:
        algorithms (List[Dict[str, str]]): a list of algorithm applications as dictionaries.
        filters (Dict[str, str]): the filters to apply.

    Returns:
        the filtered algorithms.
    """
    return [
        application
        for application in algorithms
        if all(
            [
                application[filter_type] == filter_value if filter_value else True
                for filter_type, filter_value in filters.items()
            ]
        )
    ]


def get_configuration_tuples(
    algorithms: List[Dict[str, str]]
) -> List[ConfigurationTuple]:
    """
    Returning configuration tuples from a list of applications.

    Args:
        algorithms (List[Dict[str, str]]): a list of algorithm applications as dictionaries.

    Returns:
        the configuration tuples.
    """
    return sorted(
        list(
            {
                ConfigurationTuple(
                    **{
                        key: value
                        for key, value in algorithm.items()
                        if key in ConfigurationTuple.__annotations__.keys()
                    }
                )
                for algorithm in algorithms
            }
        )
    )



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Implementation of MoLeR conditional generators."""

import logging
from itertools import cycle, islice
from typing import List

from rdkit import Chem
from molecule_generation import VaeWrapper

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class MoLeRGenerator:
    """Interface for MoLeR generator."""

    def __init__(
        self,
        resources_path: str,
        scaffolds: str,
        num_samples: int,
        beam_size: int,
        seed: int,
        num_workers: int,
    ) -> None:
        """Instantiate a MoLeR generator.

        Args:
            resources_path: path to the resources for model loading.
            scaffolds: scaffolds as '.'-separated SMILES. If empty, no scaffolds are used.
            num_samples: Number of molecules to sample per call.
            beam_size: beam size to use during decoding.
            seed: seed used for random number generation.
            num_workers: number of workers used for generation.

        Raises:
            RuntimeError: in the case extras are disabled.
        """
        # loading artifacts
        self.resources_path = resources_path
        self.scaffolds = scaffolds
        self.num_samples = num_samples
        self.beam_size = beam_size
        self.num_workers = num_workers
        self._seed = seed

    def generate(self) -> List[str]:
        """Sample molecules using MoLeR.

        Returns:
            sampled molecule (SMILES).
        """
        # process scaffolds
        valid_scaffolds = [
            scaffold
            for scaffold in self.scaffolds.split(".")
            if Chem.MolFromSmiles(scaffold) is not None
        ]
        # generate molecules
        logger.info("running MoLeR...")
        with VaeWrapper(
            self.resources_path,
            beam_size=self.beam_size,
            seed=self._seed,
            num_workers=self.num_workers,
        ) as model:
            latents = model.sample_latents(self.num_samples)
            scaffolds = list(islice(cycle(valid_scaffolds), self.num_samples))
            samples = model.decode(
                latents=latents,
                scaffolds=scaffolds if len(scaffolds) == self.num_samples else None,
            )
        # offset seed to guarantee uniqueness
        self._seed += 1
        logger.info("MoLeR run completed")
        return samples



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""RegressionTransformer tests."""

import pickle
from typing import ClassVar, Dict, Type

import numpy as np
import pytest

from gt4sd.algorithms.conditional_generation.regression_transformer import (
    RegressionTransformer,
    RegressionTransformerMolecules,
    RegressionTransformerProteins,
)
from gt4sd.algorithms.core import AlgorithmConfiguration
from gt4sd.algorithms.registry import ApplicationsRegistry
from gt4sd.tests.utils import GT4SDTestSettings

test_settings = GT4SDTestSettings.get_instance()


def get_classvar_type(class_var):
    """Extract type from ClassVar type annotation: `ClassVar[T]] -> T`."""
    return class_var.__args__[0]


@pytest.mark.parametrize(
    "config_class, algorithm_type, domain, algorithm_name",
    [
        (
            RegressionTransformerMolecules,
            "conditional_generation",
            "materials",
            RegressionTransformer.__name__,
        ),
        (
            RegressionTransformerProteins,
            "conditional_generation",
            "materials",
            RegressionTransformer.__name__,
        ),
    ],
)
def test_config_class(
    config_class: Type[AlgorithmConfiguration],
    algorithm_type: str,
    domain: str,
    algorithm_name: str,
):
    assert config_class.algorithm_type == algorithm_type
    assert config_class.domain == domain
    assert config_class.algorithm_name == algorithm_name

    for keyword, type_annotation in config_class.__annotations__.items():
        if keyword in ("algorithm_type", "domain", "algorithm_name"):
            assert type_annotation.__origin__ is ClassVar  # type: ignore
            assert str == get_classvar_type(type_annotation)


@pytest.mark.parametrize(
    "config_class",
    [
        (RegressionTransformerMolecules),
        (RegressionTransformerProteins),
    ],
)
def test_config_instance(config_class: Type[AlgorithmConfiguration]):
    config = config_class()  # type:ignore
    assert config.algorithm_application == config_class.__name__


@pytest.mark.parametrize(
    "config_class",
    [
        (RegressionTransformerMolecules),
        (RegressionTransformerProteins),
    ],
)
def test_available_versions(config_class: Type[AlgorithmConfiguration]):
    versions = config_class.list_versions()
    assert "solubility" in versions or "stability" in versions


@pytest.mark.parametrize(
    "config, example_target, algorithm, params",
    [
        pytest.param(
            RegressionTransformerMolecules,
            "<esol>[MASK][MASK][MASK][MASK][MASK]|[Cl][C][Branch1_2][Branch1_2][=C][Branch1_1][C][Cl][Cl][Cl]",
            RegressionTransformer,
            {"algorithm_version": "solubility", "search": "greedy", "num_samples": 1},
            marks=pytest.mark.skipif(test_settings.gt4sd_ci, reason="high_memory"),
        ),
        pytest.param(
            RegressionTransformerMolecules,
            "<esol>-3.49|[C][C][MASK][MASK][MASK][C][Br]",
            RegressionTransformer,
            {"search": "sample", "temperature": 2.0, "num_samples": 5},
            marks=pytest.mark.skipif(test_settings.gt4sd_ci, reason="high_memory"),
        ),
        pytest.param(
            RegressionTransformerMolecules,
            "<logp>[MASK][MASK][MASK][MASK][MASK]|<scs>[MASK][MASK][MASK][MASK][MASK]|[C][C][O][C][=N][C][=N][C][Branch1_2][Branch1_1][=C][Ring1][Branch1_2][C][N][C][C][O][C][Branch1_1][C][C][C]",
            RegressionTransformer,
            {
                "search": "greedy",
                "num_samples": 1,
                "algorithm_version": "logp_and_synthesizability",
            },
            marks=pytest.mark.skipif(test_settings.gt4sd_ci, reason="high_memory"),
        ),
        pytest.param(
            RegressionTransformerMolecules,
            "<logp>6.534|<scs>3.835|[C][MASK][MASK][C][=N][C][=N][C][Branch1_2][Branch1_1][=C][Ring1][Branch1_2][C][N][C][C][O][C][Branch1_1][C][C][C]",
            RegressionTransformer,
            {
                "search": "sample",
                "num_samples": 3,
                "temperature": 1.4,
                "algorithm_version": "logp_and_synthesizability",
            },
            marks=pytest.mark.skipif(test_settings.gt4sd_ci, reason="high_memory"),
        ),
        pytest.param(
            RegressionTransformerProteins,
            "<stab>[MASK][MASK][MASK][MASK][MASK]|GSQEVNSNASPEEAEIARKAGATTWTEKGNKWEIRI",
            RegressionTransformer,
            {"search": "greedy", "num_samples": 1},
            marks=pytest.mark.skipif(test_settings.gt4sd_ci, reason="high_memory"),
        ),
        pytest.param(
            RegressionTransformerProteins,
            "<stab>1.123|TTIKNG[MASK][MASK][MASK]YTVPLSPEQAAK[MASK][MASK][MASK]KKRWPDYEVQIHGNTVKVT",
            RegressionTransformer,
            {"search": "sample", "temperature": 2.0, "num_samples": 5},
            marks=pytest.mark.skipif(test_settings.gt4sd_ci, reason="high_memory"),
        ),
        # Test the sampling wrapper configurations
        pytest.param(
            RegressionTransformerProteins,
            "TTIKNGABCYTVPLSPEQAAKABCKKRWPDYEVQIHGNTVKVT",
            RegressionTransformer,
            {
                "search": "sample",
                "temperature": 2.0,
                "num_samples": 5,
                "tolerance": 5,
                "sampling_wrapper": {
                    "property_goal": {"<stab>": 1.123},
                    "fraction_to_mask": 0.9,
                    "tokens_to_mask": ["A"],
                },
            },
            marks=pytest.mark.skipif(test_settings.gt4sd_ci, reason="high_memory"),
        ),
        pytest.param(
            RegressionTransformerProteins,
            "TTIKNGABCYTVPLSPEQAAKABCKKRWPDYEVQIHGNTVKVT",
            RegressionTransformer,
            {
                "search": "sample",
                "temperature": 2.0,
                "num_samples": 5,
                "tolerance": 80,
                "sampling_wrapper": {
                    "property_goal": {"<stab>": 1.123},
                    "fraction_to_mask": 0.9,
                    "tokens_to_mask": ["A"],
                    "substructures_to_keep": ["KABCKKRWP"],
                    "substructures_to_mask": ["NGABCYT"],
                },
            },
            marks=pytest.mark.skipif(test_settings.gt4sd_ci, reason="high_memory"),
        ),
        pytest.param(
            RegressionTransformerMolecules,
            "CCOC1=NC=NC(=C1C)NCCOC(C)C",
            RegressionTransformer,
            {
                "search": "sample",
                "num_samples": 3,
                "temperature": 1.4,
                "tolerance": {"<logp>": 5, "<scs>": 10},
                "algorithm_version": "logp_and_synthesizability",
                "sampling_wrapper": {
                    "property_goal": {"<logp>": 6.534, "<scs>": 3.835},
                    "fraction_to_mask": 0.2,
                },
            },
            marks=pytest.mark.skipif(test_settings.gt4sd_ci, reason="high_memory"),
        ),
        pytest.param(
            RegressionTransformerMolecules,
            "CC(C#C)N(C)C(=O)NC1=CC=C(Cl)C=C1",
            RegressionTransformer,
            {
                "search": "sample",
                "num_samples": 3,
                "temperature": 1.1,
                "tolerance": 80,
                "algorithm_version": "qed",
                "sampling_wrapper": {
                    "property_goal": {"<qed>": 0.5},
                    "fraction_to_mask": 0.2,
                    "substructures_to_keep": ["C1=CC=C(Cl)C=C1"],
                    "substructures_to_mask": ["C(=O)"],
                },
            },
            marks=pytest.mark.skipif(test_settings.gt4sd_ci, reason="high_memory"),
        ),
        pytest.param(
            RegressionTransformerMolecules,
            "<[A];CC1=CC=C(CO[R])C=C1;[A].R -> [B].R><[B];O=C(O[C@H](C([R])=O)C)[C@@H](O[Q])C;[B].Q -> [B].R>",
            RegressionTransformer,
            {
                "search": "sample",
                "num_samples": 3,
                "temperature": 0.9,
                "tolerance": 100,
                "algorithm_version": "block_copolymer",
                "sampling_wrapper": {
                    "property_goal": {"<Dispersity>": 1.080, "<MnGPC>": 3513},
                    "fraction_to_mask": 0.1,
                },
            },
            marks=pytest.mark.skipif(test_settings.gt4sd_ci, reason="high_memory"),
        ),
        pytest.param(
            RegressionTransformerMolecules,
            "<Dispersity>1.080|<MnGPC>3513.00000|<[A];CC1=CC[MASK][MASK](CO[R])C=C1;[A].R -> [B].R><[B];O=C(O[C@H](C([R])=O)C)[C@@H](O[Q])C;[B].Q -> [B].R>",
            RegressionTransformer,
            {
                "algorithm_version": "block_copolymer",
                "search": "greedy",
                "num_samples": 2,
                "tolerance": 100,
            },
            marks=pytest.mark.skipif(test_settings.gt4sd_ci, reason="high_memory"),
        ),
        pytest.param(
            RegressionTransformerMolecules,
            "<Dispersity>[MASK][MASK][MASK][MASK][MASK]|<MnGPC>[MASK][MASK][MASK][MASK][MASK][MASK][MASK][MASK][MASK][MASK]|<[A];CC1=CC=C(CO[R])C=C1;[A].R -> [B].R><[B];O=C(O[C@H](C([R])=O)C)[C@@H](O[Q])C;[B].Q -> [B].R>",
            RegressionTransformer,
            {
                "algorithm_version": "block_copolymer",
                "search": "greedy",
                "num_samples": 1,
            },
            marks=pytest.mark.skipif(test_settings.gt4sd_ci, reason="high_memory"),
        ),
    ],
)
def test_generation_via_import(config, example_target, algorithm, params):
    num_samples = params.pop("num_samples", 1)
    regression_transformer = algorithm(
        configuration=config(**params), target=example_target
    )

    # Check if tolerance was propagated
    if "tolerance" in params.keys():
        if isinstance(params["tolerance"], Dict):
            np.testing.assert_almost_equal(
                list(params["tolerance"].values()),
                list(
                    regression_transformer.configuration.generator.tolerance_dict.values()
                ),
            )
        elif isinstance(params["tolerance"], float):
            np.testing.assert_almost_equal(
                list([params["tolerance"] * len(params["property_goal"].keys())]),
                list(
                    regression_transformer.configuration.generator.tolerance_dict.values()
                ),
            )

    items = list(regression_transformer.sample(num_samples))
    assert len(items) == num_samples


@pytest.mark.parametrize(
    "algorithm_application, target, params",
    [
        pytest.param(
            RegressionTransformerMolecules.__name__,
            "<esol>[MASK][MASK][MASK][MASK][MASK]|[Cl][C][Branch1_2][Branch1_2][=C][Branch1_1][C][Cl][Cl][Cl]",
            {"search": "greedy", "num_samples": 1},
            marks=pytest.mark.skipif(test_settings.gt4sd_ci, reason="high_memory"),
        ),
        pytest.param(
            RegressionTransformerMolecules.__name__,
            "<esol>-3.49|[C][C][MASK][MASK][MASK][C][Br]",
            {"search": "sample", "temperature": 2.0, "num_samples": 5},
            marks=pytest.mark.skipif(test_settings.gt4sd_ci, reason="high_memory"),
        ),
        pytest.param(
            RegressionTransformerProteins.__name__,
            "<stab>[MASK][MASK][MASK][MASK][MASK]|GSQEVNSNASPEEAEIARKAGATTWTEKGNKWEIRI",
            {"search": "greedy", "num_samples": 1},
            marks=pytest.mark.skipif(test_settings.gt4sd_ci, reason="high_memory"),
        ),
        pytest.param(
            RegressionTransformerProteins.__name__,
            "<stab>1.123|TTIKNG[MASK][MASK][MASK]YTVPLSPEQAAK[MASK][MASK][MASK]KKRWPDYEVQIHGNTVKVT",
            {"search": "sample", "temperature": 2.0, "num_samples": 5},
            marks=pytest.mark.skipif(test_settings.gt4sd_ci, reason="high_memory"),
        ),
    ],
)
def test_generation_via_registry(target, algorithm_application, params):
    num_samples = params.pop("num_samples", 1)
    regression_transformer = ApplicationsRegistry.get_application_instance(
        target=target,
        algorithm_type="conditional_generation",
        domain="materials",
        algorithm_name=RegressionTransformer.__name__,
        algorithm_application=algorithm_application,
        **params,
    )

    items = list(regression_transformer.sample(num_samples))
    assert len(items) == num_samples


@pytest.mark.parametrize(
    "config_class",
    [
        (RegressionTransformerMolecules),
        (RegressionTransformerProteins),
    ],
)
def test_configuration_pickable(config_class: Type[AlgorithmConfiguration]):
    # implementation
    obj = config_class(algorithm_version="test")

    # ---
    import inspect

    inspect.getmodule(config_class)
    # ---
    pickled_obj = pickle.dumps(obj)
    restored_obj = pickle.loads(pickled_obj)
    assert restored_obj.algorithm_version == "test"
    assert restored_obj == obj

    # registered
    Config = ApplicationsRegistry.get_application(
        algorithm_type="conditional_generation",
        domain="materials",
        algorithm_name=RegressionTransformer.__name__,
        algorithm_application=config_class.__name__,
    ).configuration_class

    obj = Config(algorithm_version="test")
    pickled_obj = pickle.dumps(obj)
    restored_obj = pickle.loads(pickled_obj)

    assert restored_obj.algorithm_version == "test"
    assert restored_obj == obj



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""TorchDrug training utilities."""
import os
from dataclasses import dataclass, field
from typing import Any, Dict, Optional

from ...configuration import gt4sd_configuration_instance
from ..core import TrainingPipeline, TrainingPipelineArguments
from . import DATASET_FACTORY

DATA_ROOT = os.path.join(
    gt4sd_configuration_instance.gt4sd_local_cache_path, "data", "torchdrug"
)
os.makedirs(DATA_ROOT, exist_ok=True)


class TorchDrugTrainingPipeline(TrainingPipeline):
    """TorchDrug training pipelines."""

    def train(  # type: ignore
        self,
        training_args: Dict[str, Any],
        model_args: Dict[str, Any],
        dataset_args: Dict[str, Any],
    ) -> None:
        """Generic training function for launching a TorchDrug training.

        Args:
            training_args: training arguments passed to the configuration.
            model_args: model arguments passed to the configuration.
            dataset_args: dataset arguments passed to the configuration.

        Raises:
            NotImplementedError: the generic trainer does not implement the pipeline.
        """
        raise NotImplementedError


@dataclass
class TorchDrugTrainingArguments(TrainingPipelineArguments):
    """Arguments related to torchDrug trainer."""

    __name__ = "training_args"

    model_path: str = field(
        metadata={"help": "Path where the model artifacts are stored."}
    )
    training_name: str = field(metadata={"help": "Name used to identify the training."})
    epochs: int = field(default=10, metadata={"help": "Number of epochs."})
    batch_size: int = field(default=16, metadata={"help": "Size of the batch."})
    learning_rate: float = field(
        default=1e-5, metadata={"help": "Learning rate used in training."}
    )
    log_interval: int = field(
        default=100, metadata={"help": "Number of steps between log intervals."}
    )
    gradient_interval: int = field(
        default=1, metadata={"help": "Gradient accumulation steps"}
    )
    num_worker: int = field(
        default=0, metadata={"help": "Number of CPU workers per GPU."}
    )
    task: Optional[str] = field(
        default=None,
        metadata={
            "help": "Optimization task for goal-driven generation."
            "Currently, TorchDrug only supports `plogp` and `qed`."
        },
    )


@dataclass
class TorchDrugDataArguments(TrainingPipelineArguments):
    """Arguments related to TorchDrug data loading."""

    __name__ = "dataset_args"

    dataset_name: str = field(
        metadata={
            "help": f"Identifier for the dataset. Has to be in {DATASET_FACTORY.keys()}"
            ". Can either point to one of the predefined TorchDrug datasets or it can "
            "be `custom` if the user brings their own dataset. If `custom`, then the "
            "arguments `file_path`, `target_field` and `smiles_field` below have to be"
            " specified."
        }
    )
    file_path: str = field(
        default="",
        metadata={
            "help": "Ignored unless `datase_name` is `custom`. In that case it's "
            "a path to a .csv file containing the training data."
        },
    )
    dataset_path: str = field(
        default=DATA_ROOT,
        metadata={
            "help": "Path where the TorchDrug dataset will be stored. This is ignored "
            "if `datase_name` is `custom`."
        },
    )
    target_field: str = field(
        default="",
        metadata={
            "help": "Ignored unless `datase_name` is `custom`. In that case it's a str "
            "with name of the column containing the property that can be optimized."
            "Currently TorchDrug only supports `plogp` and `qed`."
        },
    )
    smiles_field: str = field(
        default="smiles",
        metadata={
            "help": "Ignored unless `datase_name` is `custom`. In that case it's the "
            "name of the column containing the SMILES strings."
        },
    )
    transform: str = field(
        default="lambda x: x",
        metadata={
            "help": "Optional data transformation function. Has to be a lambda function"
            " (written as a string) that operates on the batch dictionary."
            "See torchdrug docs for details."
        },
    )
    verbose: int = field(
        default=1, metadata={"help": "Output verbosity level for dataset."}
    )
    lazy: bool = field(
        default=False,
        metadata={
            "help": "If yes, molecules are processed in the dataloader. This is faster "
            "for setup but slower at training time."
        },
    )
    node_feature: str = field(
        default="default",
        metadata={"help": "Node features (or node feature list) to extract."},
    )
    edge_feature: str = field(
        default="default",
        metadata={"help": "Edge features (or edge feature list) to extract."},
    )
    graph_feature: Optional[str] = field(
        default=None,
        metadata={"help": "Graph features (or graph feature list) to extract."},
    )
    with_hydrogen: bool = field(
        default=False,
        metadata={"help": "Whether hydrogens are stored in molecular graph."},
    )
    no_kekulization: bool = field(
        default=False,
        metadata={
            "help": "Whether SMILES kekulization is used. Per default, it is used."
        },
    )


@dataclass
class TorchDrugSavingArguments(TrainingPipelineArguments):
    """Saving arguments related to TorchDrug trainer."""

    __name__ = "saving_args"

    model_path: str = field(
        metadata={"help": "Path where the model artifacts are stored."}
    )
    training_name: str = field(metadata={"help": "Name used to identify the training."})
    dataset_name: str = field(
        metadata={
            "help": f"Identifier for the dataset. Has to be in {DATASET_FACTORY.keys()}"
            ". Can either point to one of the predefined TorchDrug datasets or it can "
            "be `custom` if the user brings their own dataset. If `custom`, then the "
            "arguments `file_path`, `target_field` and `smiles_field` below have to be"
            " specified."
        }
    )
    task: Optional[str] = field(
        default=None,
        metadata={
            "help": "Optimization task for goal-driven generation."
            "Currently, TorchDrug only supports `plogp` and `qed`."
        },
    )
    file_path: str = field(
        default="",
        metadata={
            "help": "Ignored unless `datase_name` is `custom`. In that case it's "
            "a path to a .csv file containing the training data."
        },
    )
    epochs: int = field(default=10, metadata={"help": "Number of epochs."})



# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Feature engine for crystals."""

import os
import re
from typing import Dict, List

import pandas as pd
import pkg_resources  # type: ignore
from pymatgen.core.composition import Composition


class Features:
    """Feature generator for crystals."""

    def __init__(self, formula_file: str):
        """Initialize Feature engine.

        Args:
            formula_file: file of formulas
        """
        self.formula_file = formula_file
        self.atomic_data_file = pkg_resources.resource_filename(  # type: ignore
            "gt4sd", os.path.join("frameworks", "crystals_rfc", "atomic_data.csv")
        )

    def make_features(
        self,
        atomic_descriptors,
        formula_list,
        targets,
        encoded_sym,
        add_avg: bool = True,
        add_aad: bool = True,
        add_md: bool = False,
        add_cv: bool = False,
    ) -> List[List[float]]:
        """Initialize Feature engine.

        Args:
            atomic_descriptors: atomic descriptors.
            formula_list: file of formulas.
            targets: targets.
            encoded_sym: encoded_sym.
            add_avg: include Weighted Average.
            add_aad: include Average Absolute Deviation.
            add_md: include maximum difference.
            add_cv: include element ratio vector.

        Returns:
            List of all features.
        """

        all_feature_list = []
        for indx0, formula in enumerate(formula_list):

            feature_list = []
            atom_symbols = list(atomic_descriptors[0].keys())

            comp = Composition(formula)
            formula = comp.formula
            s = re.findall(r"([A-Z][a-z]?)([0-9]?\.?[0-9]*)", formula)
            comp_vector = [0 for _ in range(0, len(atom_symbols))]

            feature_list.append(targets[indx0])

            for d in encoded_sym[indx0]:
                feature_list.append(d)

            # Calculating the total number of atoms in the chemical formula
            total = 0.0
            for elem, num in s:
                if num == "":
                    num = 1
                total += int(num)

            # Calculating Weighted Average
            avg = 0.0
            for des in atomic_descriptors:
                des_list = []
                for elem, num in s:
                    if num == "":
                        num = 1
                    num = int(num)
                    avg += des[elem] * num
                    des_list.append((des[elem], num))
                avg = avg / total
                if add_avg:
                    feature_list.append(avg)

                # Calculating Average Absolute Deviation
                if add_aad:
                    avg_ad = 0.0
                    for y, num in des_list:
                        ad = abs(y - avg) * num
                        avg_ad += ad
                    avg_ad = avg_ad / total

                    feature_list.append(avg_ad)

                # Calculating maximum difference
                if add_md:
                    dif_list = []
                    for y1, num1 in des_list:
                        for y2, num2 in des_list:
                            dif = abs(y1 - y2)
                            dif_list.append(dif)
                    max_dif = max(dif_list)

                    feature_list.append(max_dif)

            # Creting Element Ratio Vector
            for elem, num in s:
                if num == "":
                    num = 1
                num = int(num)
                index = atom_symbols.index(elem)
                comp_vector[index] = int(num) / total  # type: ignore

            # Uncomment if the element ratio vector is required
            if add_cv:
                for ratio in comp_vector:
                    feature_list.append(ratio)

            all_feature_list.append(feature_list)

        return all_feature_list

    def get_formula_list(self) -> List[str]:
        """Get formula list.

        Returns:
            List of formulas.
        """
        df_mat = pd.read_csv(self.formula_file, header=None)
        formula_list = [x[0] for x in df_mat.values.tolist()]

        return formula_list

    def get_encoded_sym(self) -> List[List[int]]:
        """Get encoded systems.

        Returns:
            List of encoded systems.
        """
        df_mat = pd.read_csv(self.formula_file, header=None)

        if len(df_mat.columns) == 3:
            sym_list = [x[2] for x in df_mat.values.tolist()]
        elif len(df_mat.columns) == 2:
            sym_list = [x[1] for x in df_mat.values.tolist()]
        else:
            raise ValueError(
                "The provided csv file should contain two or three columns."
            )

        encoded_sym = []
        for sym in sym_list:
            if sym == "monoclinic":
                digits = [1, 0, 0, 0, 0, 0, 0]
            elif sym == "triclinic":
                digits = [0, 1, 0, 0, 0, 0, 0]
            elif sym == "orthorhombic":
                digits = [0, 0, 1, 0, 0, 0, 0]
            elif sym == "trigonal":
                digits = [0, 0, 0, 1, 0, 0, 0]
            elif sym == "hexagonal":
                digits = [0, 0, 0, 0, 1, 0, 0]
            elif sym == "cubic":
                digits = [0, 0, 0, 0, 0, 1, 0]
            elif sym == "tetragonal":
                digits = [0, 0, 0, 0, 0, 0, 1]

            encoded_sym.append(digits)
        return encoded_sym

    def get_targets(self) -> List[float]:
        """Get targets.

        Returns:
            List of targets.
        """
        df_mat = pd.read_csv(self.formula_file, header=None)

        if len(df_mat.columns) == 3:
            targets = [x[1] for x in df_mat.values.tolist()]
        elif len(df_mat.columns) == 2:
            targets = [0 for _ in df_mat.values.tolist()]
        else:
            raise ValueError(
                "The provided csv file should contain two or three columns."
            )

        return targets

    def get_atomic_descriptors(self) -> List[Dict[str, float]]:
        """Get atomic descriptors.

        Returns:
               List of atomic descriptors.
        """
        df_des = pd.read_csv(self.atomic_data_file, header=None)

        atomic_descriptors = []

        elements = [x[0] for x in df_des.values.tolist()]
        for i in range(1, len(df_des.columns)):
            tmp = [x[i] for x in df_des.values.tolist()]
            des_dict = dict(zip(elements, tmp))

            atomic_descriptors.append(des_dict)

        return atomic_descriptors

    def get_features(self) -> List[List[float]]:
        """Get features.

        Returns:
            List of all features.
        """
        formula_list = self.get_formula_list()
        encoded_sym = self.get_encoded_sym()
        targets = self.get_targets()
        atomic_descriptors = self.get_atomic_descriptors()

        features = self.make_features(
            atomic_descriptors=atomic_descriptors,
            encoded_sym=encoded_sym,
            targets=targets,
            formula_list=formula_list,
        )

        return features



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Generic utils for pytorch."""

from typing import Dict, List, Optional, Union

import torch


def get_gpu_device_names() -> List[str]:
    """Get GPU device names as a list.

    Returns:
        names of available GPU devices.
    """
    gpu_device_names = []
    if torch.cuda.is_available():
        gpu_device_names = [
            f"cuda:{index}" for index in range(torch.cuda.device_count())
        ]
    return gpu_device_names


def claim_device_name() -> str:
    """Claim a device name.

    Returns:
        device name, if on GPU is available returns CPU.
    """
    device_name = "cpu"
    gpu_device_names = get_gpu_device_names()
    if len(gpu_device_names) > 0:
        device_name = gpu_device_names[0]
    return device_name


def get_device() -> torch.device:
    """
    Get device dynamically.
    """
    return torch.device("cuda" if torch.cuda.is_available() else "cpu")


def device_claim(device: Optional[Union[torch.device, str]] = None) -> torch.device:
    """
    Satidfy a device claim.

    Args:
        device: device where the inference is running either as a dedicated class or
            a string. If not provided is inferred.

    Returns:
        torch.device: the claimed device or a default one.
    """
    if isinstance(device, str):
        device = torch.device(device)
    device = (
        get_device()
        if (device is None or not isinstance(device, torch.device))
        else device
    )
    return device


def get_device_from_tensor(tensor: torch.Tensor) -> torch.device:
    """Get the device from a tensor.

    Args:
        tensor: a tensor.

    Returns:
        the device.
    """
    device_id = tensor.get_device()
    device = "cpu" if device_id < 0 else f"cuda:{device_id}"
    return device_claim(device)


def map_tensor_dict(
    tensor_dict: Dict[str, torch.Tensor], device: torch.device
) -> Dict[str, torch.Tensor]:
    """
    Maps a dictionary of tensors to a specific device.

    Args:
        tensor_dict: A dictionary of tensors.
        device: The device to map the tensors to.

    Returns:
        A dictionary of tensors mapped to the device.
    """
    return {key: tensor.to(device) for key, tensor in tensor_dict.items()}



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Implementation details for a Template algorithm"""

import random
from typing import List


class Generator:
    """Basic Generator for the template algorithm"""

    def __init__(self, resources_path: str, temperature: int) -> None:
        """Initialize the Generator.

        Args:
            resources_path: directory where to find models and parameters.

        """

        self.resources_path = resources_path
        self.temperature = temperature

    def hello_name(
        self,
        name: str,
    ) -> List[str]:
        """Validate a list of strings.

        Args:
            name: a string.

        Returns:
            a list containing salutation and temperature converted to fahrenheit.
        """
        return [
            f"Hello {str(name)} {random.randint(1, int(1e6))} times and, fun fact, {str(self.temperature)} celsius equals to {(self.temperature * (9/5) + 32)} fahrenheit."
        ]



#!/usr/bin/env python
#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""Run model upload for the GT4SD.
   Two steps procedure: check if the folder/model name is already in the database.
   If not, upload it.
"""

import logging
import os
import sys
from dataclasses import dataclass, field
from typing import IO, Iterable, Optional, cast

from ..algorithms.registry import ApplicationsRegistry
from ..training_pipelines import TRAINING_PIPELINE_ARGUMENTS_FOR_MODEL_SAVING
from ..training_pipelines.core import TrainingPipelineArguments
from .algorithms import (
    AVAILABLE_ALGORITHMS,
    AVAILABLE_ALGORITHMS_CATEGORIES,
    filter_algorithm_applications,
    get_configuration_tuples,
)
from .argument_parser import ArgumentParser, DataClassType

logger = logging.getLogger(__name__)

SUPPORTED_TRAINING_PIPELINES = sorted(
    TRAINING_PIPELINE_ARGUMENTS_FOR_MODEL_SAVING.keys()
)


@dataclass
class SavingArguments:
    """Algorithm saving arguments."""

    __name__ = "saving_base_args"

    training_pipeline_name: str = field(
        metadata={
            "help": f"Training pipeline name, supported pipelines: {', '.join(SUPPORTED_TRAINING_PIPELINES)}."
        },
    )
    target_version: str = field(
        metadata={"help": "Target algorithm version to save."},
    )
    algorithm_type: Optional[str] = field(
        default=None,
        metadata={
            "help": f"Inference algorithm type, supported types: {', '.join(AVAILABLE_ALGORITHMS_CATEGORIES['algorithm_type'])}."
        },
    )
    domain: Optional[str] = field(
        default=None,
        metadata={
            "help": f"Domain of the inference algorithm, supported types: {', '.join(AVAILABLE_ALGORITHMS_CATEGORIES['domain'])}."
        },
    )
    algorithm_name: Optional[str] = field(
        default=None,
        metadata={"help": "Inference algorithm name."},
    )
    algorithm_application: Optional[str] = field(
        default=None,
        metadata={"help": "Inference algorithm application."},
    )
    source_version: Optional[str] = field(
        default=None,
        metadata={"help": "Source algorithm version to use for missing artifacts."},
    )


class SavingArgumentParser(ArgumentParser):
    """Argument parser using a custom help logic."""

    def print_help(self, file: Optional[IO[str]] = None) -> None:
        """Print help checking dynamically whether a specific pipeline is passed.

        Args:
            file: an optional I/O stream. Defaults to None, a.k.a., stdout and stderr.
        """
        try:
            help_args_set = {"-h", "--help"}
            if (
                len(set(sys.argv).union(help_args_set)) < len(help_args_set) + 2
            ):  # considering filename
                super().print_help()
                return
            args = [arg for arg in sys.argv if arg not in help_args_set]
            parsed_arguments = super().parse_args_into_dataclasses(
                args=args, return_remaining_strings=True
            )
            trainer_arguments = None
            for arguments in parsed_arguments:
                if arguments.__name__ == "trainer_base_args":
                    trainer_arguments = arguments
                    break
            if trainer_arguments:
                trainer_arguments.training_pipeline_name
                training_pipeline_arguments = (
                    TRAINING_PIPELINE_ARGUMENTS_FOR_MODEL_SAVING.get(
                        trainer_arguments.training_pipeline_name,
                        TrainingPipelineArguments,
                    )
                )
                parser = ArgumentParser(
                    tuple(
                        [SavingArguments, *training_pipeline_arguments]  # type:ignore
                    )
                )
                parser.print_help()
        except Exception:
            super().print_help()


def main() -> None:
    """
    Run an algorithm saving pipeline.

    Raises:
        ValueError: in case the provided training pipeline provided is not supported.
    """
    logging.basicConfig(stream=sys.stdout, level=logging.INFO)

    base_args = SavingArgumentParser(
        cast(DataClassType, SavingArguments)
    ).parse_args_into_dataclasses(return_remaining_strings=True)[0]
    training_pipeline_name = base_args.training_pipeline_name

    if training_pipeline_name not in set(SUPPORTED_TRAINING_PIPELINES):
        ValueError(
            f"Training pipeline {training_pipeline_name} is not supported. Supported types: {', '.join(SUPPORTED_TRAINING_PIPELINES)}."
        )

    training_pipeline_saving_arguments = TRAINING_PIPELINE_ARGUMENTS_FOR_MODEL_SAVING[
        training_pipeline_name
    ]
    parser = SavingArgumentParser(
        cast(
            Iterable[DataClassType],
            tuple([SavingArguments, training_pipeline_saving_arguments]),
        )
    )

    saving_args, training_pipeline_saving_args, _ = parser.parse_args_into_dataclasses(
        return_remaining_strings=True
    )

    filters = {
        key: saving_args.__dict__[key]
        for key in [
            "algorithm_type",
            "algorithm_application",
            "domain",
            "algorithm_name",
            "source_version",
        ]
    }
    configuration_tuples = get_configuration_tuples(
        filter_algorithm_applications(algorithms=AVAILABLE_ALGORITHMS, filters=filters)
    )

    # too many configurations compatible
    if len(configuration_tuples) > 1:
        logger.info(
            f"Multiple configurations matching the parameters:{os.linesep}"
            f"{os.linesep.join(map(str, configuration_tuples))}{os.linesep}"
            f"Select one by specifying additional algorithms parameters: {','.join('--' + key for key, value in filters.items() if not value)}.",
        )
        return

    # no configurations compatible
    elif len(configuration_tuples) < 1:
        provided_filters = {key: value for key, value in filters.items() if value}
        logger.error(
            "No configurations matching the provided parameters, "
            f"please review the supported configurations:{os.linesep}"
            f"{os.linesep.join(map(str, configuration_tuples))}{os.linesep}"
            f"Please review the parameters provided:{os.linesep}"
            f"{provided_filters}"
        )
    configuration_tuple = configuration_tuples[0]
    logger.info(f"Selected configuration: {configuration_tuple}")

    algorithm_application = ApplicationsRegistry.applications[configuration_tuple]
    configuration_class = algorithm_application.configuration_class
    logger.info(
        f'Saving model version "{saving_args.target_version}" with the following configuration: {configuration_class}'
    )

    # implement upload method using minio and s3.py
    # core.py and configuration.py
    configuration_class.upload_version_from_training_pipeline_arguments(
        training_pipeline_arguments=training_pipeline_saving_args,
        target_version=saving_args.target_version,
        source_version=saving_args.source_version,
    )


if __name__ == "__main__":
    main()



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""SMILES LSTM training pipeline from GuacaMol."""
import logging
import os
from dataclasses import dataclass, field
from typing import Any, Dict

from guacamol_baselines.smiles_lstm_hc.smiles_rnn_distribution_learner import (
    SmilesRnnDistributionLearner,
)

from ...core import TrainingPipelineArguments
from ..core import GuacaMolBaselinesTrainingPipeline

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class GuacaMolLSTMTrainingPipeline(GuacaMolBaselinesTrainingPipeline):
    """GuacaMol SMILES LSTM training pipeline."""

    def train(  # type: ignore
        self,
        training_args: Dict[str, Any],
        model_args: Dict[str, Any],
        dataset_args: Dict[str, Any],
    ) -> None:
        params = {**training_args, **model_args, **dataset_args}

        os.makedirs(params["output_dir"], exist_ok=True)

        with open(params.pop("train_smiles_filepath")) as f:
            train_list = f.readlines()

        with open(params.pop("test_smiles_filepath")) as f:
            valid_list = f.readlines()
        trainer = SmilesRnnDistributionLearner(**params)

        trainer.train(training_set=train_list, validation_set=valid_list)


@dataclass
class GuacaMolLSTMTrainingArguments(TrainingPipelineArguments):
    """Training Arguments related to SMILES LSTM trainer."""

    __name__ = "training_args"

    output_dir: str = field(metadata={"help": "Output directory."})
    batch_size: int = field(
        default=512, metadata={"help": "Size of a mini-batch for gradient descent."}
    )
    valid_every: int = field(
        default=1000, metadata={"help": "Validate every so many batches."}
    )
    n_epochs: int = field(default=10, metadata={"help": "Number of training epochs."})
    lr: float = field(default=1e-3, metadata={"help": "RNN learning rate."})


@dataclass
class GuacaMolLSTMModelArguments(TrainingPipelineArguments):
    """Arguments related to SMILES LSTM trainer."""

    __name__ = "model_args"

    hidden_size: int = field(default=512, metadata={"help": "Size of hidden layer."})
    n_layers: int = field(
        default=3, metadata={"help": "Number of layers for training."}
    )
    rnn_dropout: float = field(default=0.2, metadata={"help": "Dropout value for RNN."})
    max_len: int = field(
        default=100, metadata={"help": "Max length of a SMILES string."}
    )



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Functions to facilitate conversion from dataclasses to training descriptions."""

import logging
from dataclasses import _MISSING_TYPE, fields
from typing import Any, Dict, Optional, Type, Union

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


def find_type(input_type: Type) -> Optional[str]:
    """Convert type class to string.

    Args:
        input_type: type to be converted to string.

    Returns:
        string of the type or None in case the given type is not supported.
    """
    field_type = None
    if input_type is str:
        field_type = "string"
    elif input_type is int:
        field_type = "integer"
    elif input_type is float:
        field_type = "number"
    elif input_type is bool:
        field_type = "boolean"

    return field_type


def extract_fields_from_class(
    dataclass: Type,
) -> Dict[str, Any]:
    """Extract arguments from dataclass.

    Args:
        dataclass: Dataclass to contains the arguments.

    Returns:
        Dictionary of the existing arguments including their type, description and default value.
    """

    # assign type and description
    arg_fields = {
        field.name: {
            "type": field.type,
            "description": field.metadata.get("help", "No help provided"),
        }
        for field in fields(dataclass)
    }

    # assign default values
    for field in fields(dataclass):

        if not isinstance(field.default, _MISSING_TYPE):

            if field.default is None:
                field.default = "none"

            arg_fields[field.name]["default"] = field.default

    # convert type to str
    for field_name in arg_fields:

        field_type = find_type(arg_fields[field_name]["type"])

        if field_type:

            arg_fields[field_name]["type"] = field_type

        elif (
            hasattr(arg_fields[field_name]["type"], "__origin__")
            and arg_fields[field_name]["type"].__origin__ is Union
        ):

            types = [
                find_type(type) for type in arg_fields[field_name]["type"].__args__
            ]
            types = [type for type in types if type is not None]

            if len(types) == 1:
                arg_fields[field_name]["type"] = types[0]
            else:
                raise ValueError(f"{arg_fields[field_name]['type']} not supported")

        else:
            # NOTE: not raising since the HF training args might introduce typing inconsistencies
            # Could be changed once the following is merged: https://github.com/huggingface/transformers/pull/17934
            logger.error(
                f" argument {field_name}: {arg_fields[field_name]['type']} not supported"
            )

    return arg_fields



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
from typing import Dict, Tuple, Type, Union

from ...algorithms.core import PredictorAlgorithm
from ..core import PropertyPredictor, PropertyPredictorParameters
from .core import (
    ActivityAgainstTarget,
    ActivityAgainstTargetParameters,
    Askcos,
    AskcosParameters,
    Bertz,
    ClinTox,
    ClinToxParameters,
    Docking,
    DockingParameters,
    DockingTdc,
    DockingTdcParameters,
    Esol,
    IsScaffold,
    Lipinski,
    Logp,
    MolecularWeight,
    MoleculeOne,
    MoleculeOneParameters,
    NumberAromaticRings,
    NumberAtoms,
    NumberHAcceptors,
    NumberHDonors,
    NumberHeterocycles,
    NumberLargeRings,
    NumberRings,
    NumberRotatableBonds,
    NumberStereocenters,
    OrganTox,
    OrganToxParameters,
    Plogp,
    Qed,
    Sas,
    Scscore,
    ScscoreConfiguration,
    Sider,
    SiderParameters,
    SimilaritySeed,
    SimilaritySeedParameters,
    Tox21,
    Tox21Parameters,
    Tpsa,
)

# NOTE: all functions can be called with either a SMILES or a rdkit.Chem.Mol object.
MOLECULE_PROPERTY_PREDICTOR_FACTORY: Dict[
    str,
    Tuple[
        Union[Type[PropertyPredictor], Type[PredictorAlgorithm]],
        Type[PropertyPredictorParameters],
    ],
] = {
    # inherent properties
    "molecular_weight": (MolecularWeight, PropertyPredictorParameters),
    "number_of_aromatic_rings": (NumberAromaticRings, PropertyPredictorParameters),
    "number_of_h_acceptors": (NumberHAcceptors, PropertyPredictorParameters),
    "number_of_h_donors": (NumberHDonors, PropertyPredictorParameters),
    "number_of_atoms": (NumberAtoms, PropertyPredictorParameters),
    "number_of_rings": (NumberRings, PropertyPredictorParameters),
    "number_of_rotatable_bonds": (NumberRotatableBonds, PropertyPredictorParameters),
    "number_of_large_rings": (NumberLargeRings, PropertyPredictorParameters),
    "number_of_heterocycles": (NumberHeterocycles, PropertyPredictorParameters),
    "number_of_stereocenters": (NumberStereocenters, PropertyPredictorParameters),
    "is_scaffold": (IsScaffold, PropertyPredictorParameters),
    # rule-based properties
    "bertz": (Bertz, PropertyPredictorParameters),
    "tpsa": (Tpsa, PropertyPredictorParameters),
    "logp": (Logp, PropertyPredictorParameters),
    "qed": (Qed, PropertyPredictorParameters),
    "plogp": (Plogp, PropertyPredictorParameters),
    "penalized_logp": (Plogp, PropertyPredictorParameters),
    "lipinski": (Lipinski, PropertyPredictorParameters),
    "sas": (Sas, PropertyPredictorParameters),
    "esol": (Esol, PropertyPredictorParameters),
    "similarity_seed": (SimilaritySeed, SimilaritySeedParameters),
    # properties predicted by models
    "scscore": (Scscore, ScscoreConfiguration),
    "activity_against_target": (ActivityAgainstTarget, ActivityAgainstTargetParameters),
    "tox21": (Tox21, Tox21Parameters),
    "sider": (Sider, SiderParameters),
    "organtox": (OrganTox, OrganToxParameters),
    "clintox": (ClinTox, ClinToxParameters),
    # # properties from models requiring authentification
    "askcos": (Askcos, AskcosParameters),
    "molecule_one": (MoleculeOne, MoleculeOneParameters),
    # # properties from models require additional installations
    "docking_tdc": (DockingTdc, DockingTdcParameters),
    "docking": (Docking, DockingParameters),
}


AVAILABLE_MOLECULES_PROPERTY_PREDICTOR = sorted(
    MOLECULE_PROPERTY_PREDICTOR_FACTORY.keys()
)



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Model module."""

import logging
import os
import pickle
from typing import List, Tuple

import numpy as np
import pandas as pd
from sklearn.ensemble import RandomForestClassifier
from sklearn.impute import SimpleImputer
from sklearn.model_selection import cross_val_score, train_test_split

from .feature_engine import Features

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class RFC:
    """RandomForest classifier for crystals."""

    def __init__(self, crystal_sys: str = "all"):
        """Initialize RandomForest classifier.

        Args:
            crystal_sys: crystal systems to be used.

                "all" for all the crystal systems.
                Other seven options are:
                "monoclinic", "triclinic", "orthorhombic", "trigonal",
                "hexagonal", "cubic", "tetragonal"
        """

        self.crystal_sys = crystal_sys

        self.model = RandomForestClassifier(
            n_estimators=500,
            min_samples_split=10,
            min_samples_leaf=3,
            max_features="auto",
            max_depth=70,
            bootstrap=False,
        )

        self.maxm: np.ndarray

    def load_data(self, file_name: str) -> pd.DataFrame:
        """Load dataset.

        Args:
            file_name: path of the dataset.

        Returns:
            Dataframe with the loaded dataset.
        """
        feature_eng = Features(formula_file=file_name)
        features = feature_eng.get_features()
        df = pd.DataFrame(features)
        df = df.dropna()

        imp = SimpleImputer(missing_values=np.nan, strategy="mean")
        imp.fit(df.values.tolist())
        data_list = imp.transform(df.values.tolist())
        df = pd.DataFrame(data_list)

        if self.crystal_sys == "monoclinic":
            # Selecting only the monoclinic materials
            df = df.loc[df[1] == 1]
        elif self.crystal_sys == "triclinic":
            # Selecting only the triclinic materials
            df = df.loc[df[2] == 1]
        elif self.crystal_sys == "orthorhombic":
            # Selecting only the orthorhombic materials
            df = df.loc[df[3] == 1]
        elif self.crystal_sys == "trigonal":
            # Selecting only the trigonal materials
            df = df.loc[df[4] == 1]
        elif self.crystal_sys == "hexagonal":
            # Selecting only the hexagonal materials
            df = df.loc[df[5] == 1]
        elif self.crystal_sys == "cubic":
            # Selecting only the cubic materials
            df = df.loc[df[6] == 1]
        elif self.crystal_sys == "tetragonal":
            # Selecting only the tetragonal materials
            df = df.loc[df[7] == 1]
        elif self.crystal_sys == "all":
            # Selecting  the all materials
            pass

        return df

    def split_data(
        self, df: pd.DataFrame, test_size: float = 0.2
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """Load dataset.

        Args:
            df: dataset's dataframe.
            test_size: size of the test set.

        Returns:
            Training and testing sets.
        """
        data_X, data_y = df.iloc[:, 1:].values, df.iloc[:, 0].values

        train_x, test_x, train_y, test_y = train_test_split(
            data_X, data_y, test_size=test_size, random_state=42
        )

        return train_x, test_x, train_y, test_y

    def normalize_data(
        self,
        train_x: np.ndarray,
        test_x: np.ndarray,
        train_y: np.ndarray,
        test_y: np.ndarray,
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """Normalize dataset.

        Args:
            train_x: training set's input.
            test_x: testing set's input.
            train_y: training set's groundtruth.
            test_y: testing set's groundtruth.

        Returns:
            Training and testing sets.
        """
        xx = abs(train_x)
        self.maxm = xx.max(axis=0)
        self.maxm[self.maxm == 0.0] = 1

        train_x /= self.maxm
        test_x /= self.maxm

        return train_x, test_x, train_y, test_y

    def train(self, x: np.ndarray, y: np.ndarray) -> RandomForestClassifier:
        """Train a RandomForest model.

        Args:
            x: training set's input.
            y: training set's groundtruth.

        Returns:
            Trained model.
        """

        if self.maxm is None:
            raise ValueError("Dataset should be normalized before the training.")

        scores = cross_val_score(self.model, x, y, cv=10, scoring="accuracy")
        model = self.model.fit(x, y)

        logger.info("Mean Accuracy: %0.5f (+/- %0.5f)" % (scores.mean(), scores.std()))

        return model

    def save(self, path: str) -> None:
        """Save model.

        Args:
            path: path to store the model.
        """

        if not os.path.exists(path):
            os.makedirs(path)

        # save the model to disk
        pickle.dump(self.model, open(os.path.join(path, "model.sav"), "wb"))
        # save the normalizing parameters
        df_maxm = pd.DataFrame(self.maxm)
        df_maxm.to_csv(os.path.join(path, "maxm.csv"), index=False, header=None)

    def load_model(self, path: str) -> None:
        """Save model.

        Args:
            path: path where the file is located.
        """
        # load the model from disk
        self.model = pickle.load(open(os.path.join(path, "model.sav"), "rb"))

        # load normalizing parameters
        df_maxm_load = pd.read_csv(os.path.join(path, "maxm.csv"), header=None)
        self.maxm = np.array([x[0] for x in df_maxm_load.values.tolist()])

    def predict(self, pred_x: np.ndarray) -> List[str]:
        """Predict.

        Args:
            pred_x: input.

        Returns:
            Predictions
        """

        if self.maxm is None:
            raise ValueError("Model is not initialized.")

        pred_x /= self.maxm
        y_rbf_pred = self.model.predict(pred_x)
        y_rbf_pred = list(y_rbf_pred)

        y_pred_label = ["metal" if x == 0 else "non-metal" for x in y_rbf_pred]

        return y_pred_label



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""PaccMann vanilla generator trained on polymer building blocks (catalysts/monomers)."""

import logging
import os
from dataclasses import field
from typing import ClassVar, Dict, Optional, TypeVar

from ....domains.materials import SMILES, MoleculeFormat, validate_molecules
from ....exceptions import InvalidItem
from ....training_pipelines.core import TrainingPipelineArguments
from ....training_pipelines.paccmann.core import PaccMannSavingArguments
from ...core import AlgorithmConfiguration, GeneratorAlgorithm, Untargeted
from ...registry import ApplicationsRegistry
from .implementation import Generator

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

T = type(None)
S = TypeVar("S", bound=SMILES)


class PolymerBlocks(GeneratorAlgorithm[S, T]):
    def __init__(
        self, configuration: AlgorithmConfiguration, target: Optional[T] = None
    ):
        """Polymer blocks generation.

        Args:
            configuration: domain and application
                specification, defining types and validations.
            target: unused since it is not a conditional generator.

        Example:
            An example for generating small molecules (SMILES) that resembles
            monomers/catalysts for polymer synthesis::

                configuration = PolymerBlocksGenerator()
                polymer_blocks = PolymerBlocks(configuration=configuration)
                items = list(polymer_blocks.sample(10))
                print(items)
        """

        configuration = self.validate_configuration(configuration)
        # TODO there might also be a validation/check on the target input

        super().__init__(
            configuration=configuration,
            target=None,  # type:ignore
        )

    def get_generator(
        self,
        configuration: AlgorithmConfiguration[S, T],
        target: Optional[T],
    ) -> Untargeted:
        """Get the function to sample batches via the Generator.

        Args:
            configuration: helps to set up the application.
            target: context or condition for the generation. Unused in the algorithm.

        Returns:
            callable generating a batch of items.
        """
        logger.info("ensure artifacts for the application are present.")
        self.local_artifacts = configuration.ensure_artifacts()
        implementation: Generator = configuration.get_conditional_generator(  # type: ignore
            self.local_artifacts
        )
        return implementation.sample

    def validate_configuration(
        self, configuration: AlgorithmConfiguration
    ) -> AlgorithmConfiguration:
        # TODO raise InvalidAlgorithmConfiguration
        assert isinstance(configuration, AlgorithmConfiguration)
        return configuration


@ApplicationsRegistry.register_algorithm_application(PolymerBlocks)
class PolymerBlocksGenerator(AlgorithmConfiguration[SMILES, None]):
    """Configuration to generate subunits of polymers."""

    algorithm_type: ClassVar[str] = "generation"
    domain: ClassVar[str] = "materials"
    algorithm_version: str = "v0"

    batch_size: int = field(
        default=32,
        metadata=dict(description="Batch size used for the generative model sampling."),
    )
    generated_length: int = field(
        default=100,
        metadata=dict(
            description="Maximum length in tokens of the generated molcules (relates to the SMILES length)."
        ),
    )

    def get_target_description(self) -> Optional[Dict[str, str]]:
        """Get description of the target for generation.

        Returns:
            target description, returns None in case no target is used.
        """
        return None

    def get_conditional_generator(self, resources_path: str) -> Generator:
        return Generator(
            resources_path=resources_path,
            generated_length=self.generated_length,
            batch_size=self.batch_size,
        )

    def validate_item(self, item: str) -> SMILES:
        (
            molecules,
            _,
        ) = validate_molecules([item], MoleculeFormat.smiles)
        if molecules[0] is None:
            raise InvalidItem(
                title="InvalidSMILES",
                detail=f'rdkit.Chem.MolFromSmiles returned None for "{item}"',
            )
        return SMILES(item)

    @classmethod
    def get_filepath_mappings_for_training_pipeline_arguments(
        cls, training_pipeline_arguments: TrainingPipelineArguments
    ) -> Dict[str, str]:
        """Ger filepath mappings for the given training pipeline arguments.

        Args:
            training_pipeline_arguments: training pipeline arguments.

        Returns:
            a mapping between artifacts' files and training pipeline's output files.
        """
        if isinstance(training_pipeline_arguments, PaccMannSavingArguments):
            return {
                "smiles_language.pkl": os.path.join(
                    training_pipeline_arguments.model_path,
                    f"{training_pipeline_arguments.training_name}.lang",
                ),
                "params.json": os.path.join(
                    training_pipeline_arguments.model_path,
                    training_pipeline_arguments.training_name,
                    "model_params.json",
                ),
                "weights.pt": os.path.join(
                    training_pipeline_arguments.model_path,
                    training_pipeline_arguments.training_name,
                    "weights",
                    "best_rec.pt",
                ),
            }
        else:
            return super().get_filepath_mappings_for_training_pipeline_arguments(
                training_pipeline_arguments
            )



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Test for properties."""
import numpy as np
import pytest

from gt4sd.properties import (
    MOLECULE_PROPERTY_PREDICTOR_FACTORY,
    PROTEIN_PROPERTY_PREDICTOR_FACTORY,
    SCORING_FACTORY_WITH_PROPERTY_PREDICTORS,
    PropertyPredictorRegistry,
)

protein = "KFLIYQMECSTMIFGL"
molecule = "C1=CC(=CC(=C1)Br)CN"
seed = "CCO"
_target = "drd2"
ground_truths = {
    "length": 16,
    "protein_weight": 1924.36,
    "boman_index": -0.534375,
    "aliphaticity": 97.5,
    "hydrophobicity": 0.5625,
    "charge": -0.071,
    "charge_density": -3.6895383400195386e-05,
    "isoelectric_point": 6.125,
    "aromaticity": 0.1875,
    "instability": 36.8375,
    "plogp": 0.25130060815905964,
    "molecular_weight": 186.05199999999996,
    "lipinski": 1,
    "esol": -2.6649954522215555,
    "scscore": 4.391860681753299,
    "sas": 1.6564993918409403,
    "bertz": 197.86256853719752,
    "tpsa": 26.02,
    "logp": 1.9078,
    "qed": 0.7116306772337652,
    "number_of_h_acceptors": 1,
    "number_of_atoms": 17,
    "number_of_h_donors": 1,
    "number_of_aromatic_rings": 1,
    "number_of_rings": 1,
    "number_of_large_rings": 0,
    "number_of_rotatable_bonds": 1,
    "is_scaffold": 0,
    "number_of_heterocycles": 0,
    "number_of_stereocenters": 0,
}

molecule_further_ground_truths = {
    "activity_against_target": 0.0049787821116125345,
    "similarity_seed": 0.03333333333333333,
}
protein_further_ground_truths = {"charge": 1.123}


OPTIMAL_SCORE = 1.0


def select_sample(property_key):
    """select a molecule or protein accordingly to the property."""
    return protein if property_key in PROTEIN_PROPERTY_PREDICTOR_FACTORY else molecule


def select_opposite_sample(property_key):
    """select a molecule or protein not accordingly to the property."""
    return (
        protein if property_key not in PROTEIN_PROPERTY_PREDICTOR_FACTORY else molecule
    )


@pytest.mark.parametrize(
    "property_key", [(property_key) for property_key in ground_truths.keys()]
)
def test_property_scorer(property_key):
    scoring_function = PropertyPredictorRegistry.get_property_predictor(
        name=property_key
    )
    scorer = SCORING_FACTORY_WITH_PROPERTY_PREDICTORS["property_predictor_scorer"](
        name=property_key,
        scoring_function=scoring_function,
        target=ground_truths[property_key],
    )
    sample = select_sample(property_key)
    assert np.isclose(scorer.score(sample), OPTIMAL_SCORE, atol=1e-2)  # type: ignore


def test_similarity_seed_scorer():
    scoring_function = PropertyPredictorRegistry.get_property_predictor(
        name="similarity_seed", parameters={"smiles": seed}
    )
    scorer = SCORING_FACTORY_WITH_PROPERTY_PREDICTORS["property_predictor_scorer"](
        name="similarity_seed",
        scoring_function=scoring_function,
        target=molecule_further_ground_truths["similarity_seed"],
    )
    assert np.isclose(scorer.score(molecule), OPTIMAL_SCORE, atol=1e-2)  # type: ignore


def test_activity_against_target_scorer():
    """target for the scorer is a number (for example the predicted property value). Target for the property is a string (molecule)."""
    scoring_function = PropertyPredictorRegistry.get_property_predictor(
        name="activity_against_target", parameters={"target": _target}
    )
    scorer = SCORING_FACTORY_WITH_PROPERTY_PREDICTORS["property_predictor_scorer"](
        name="activity_against_target",
        scoring_function=scoring_function,
        target=molecule_further_ground_truths["activity_against_target"],
    )
    assert np.isclose(scorer.score(molecule), OPTIMAL_SCORE, atol=1e-2)  # type: ignore


def test_charge_with_arguments_scorer():
    scoring_function = PropertyPredictorRegistry.get_property_predictor(
        name="charge", parameters={"amide": True, "ph": 5.0}
    )
    scorer = SCORING_FACTORY_WITH_PROPERTY_PREDICTORS["property_predictor_scorer"](
        name="charge",
        scoring_function=scoring_function,
        target=protein_further_ground_truths["charge"],
    )
    assert np.isclose(scorer.score(protein), OPTIMAL_SCORE, atol=1e-2)  # type: ignore


@pytest.mark.parametrize(
    "property_key", [(property_key) for property_key in ground_truths.keys()]
)
def test_validation_property_scorer(property_key):
    scoring_function = PropertyPredictorRegistry.get_property_predictor(
        name=property_key
    )
    scorer = SCORING_FACTORY_WITH_PROPERTY_PREDICTORS["property_predictor_scorer"](
        name=property_key,
        scoring_function=scoring_function,
        target=ground_truths[property_key],
    )
    sample = select_opposite_sample(property_key)
    try:
        scorer.score(sample)
        assert False
    except ValueError:
        assert True


@pytest.mark.parametrize(
    "property_key", [(property_key) for property_key in ground_truths.keys()]
)
def test_molecule_property_scorer(property_key):
    scoring_function = PropertyPredictorRegistry.get_property_predictor(
        name=property_key
    )
    if property_key not in MOLECULE_PROPERTY_PREDICTOR_FACTORY:
        try:
            scorer = SCORING_FACTORY_WITH_PROPERTY_PREDICTORS[
                "molecule_property_predictor_scorer"
            ](
                name=property_key,
                scoring_function=scoring_function,
                target=ground_truths[property_key],
            )
            assert False
        except ValueError:
            assert True

    if property_key in MOLECULE_PROPERTY_PREDICTOR_FACTORY:
        scorer = SCORING_FACTORY_WITH_PROPERTY_PREDICTORS[
            "molecule_property_predictor_scorer"
        ](
            name=property_key,
            scoring_function=scoring_function,
            target=ground_truths[property_key],
        )
        sample = select_sample(property_key)
        assert np.isclose(scorer.score(sample), OPTIMAL_SCORE, atol=1e-2)  # type: ignore


@pytest.mark.parametrize(
    "property_key", [(property_key) for property_key in ground_truths.keys()]
)
def test_protein_property_scorer(property_key):
    scoring_function = PropertyPredictorRegistry.get_property_predictor(
        name=property_key
    )
    if property_key not in PROTEIN_PROPERTY_PREDICTOR_FACTORY:
        try:
            scorer = SCORING_FACTORY_WITH_PROPERTY_PREDICTORS[
                "protein_property_predictor_scorer"
            ](
                name=property_key,
                scoring_function=scoring_function,
                target=ground_truths[property_key],
            )
            assert False
        except ValueError:
            assert True

    if property_key in PROTEIN_PROPERTY_PREDICTOR_FACTORY:
        scorer = SCORING_FACTORY_WITH_PROPERTY_PREDICTORS[
            "protein_property_predictor_scorer"
        ](
            name=property_key,
            scoring_function=scoring_function,
            target=ground_truths[property_key],
        )
        sample = select_sample(property_key)
        assert np.isclose(scorer.score(sample), OPTIMAL_SCORE, atol=1e-2)  # type: ignore


def test_property_predictor_scorer_registry():
    scorer = PropertyPredictorRegistry.get_property_predictor_scorer(
        property_name="similarity_seed",
        scorer_name="property_predictor_scorer",
        target=molecule_further_ground_truths["similarity_seed"],
        parameters={"smiles": seed},
    )
    sample = select_sample("similarity_seed")
    assert isinstance(
        scorer, SCORING_FACTORY_WITH_PROPERTY_PREDICTORS["property_predictor_scorer"]
    )
    assert np.isclose(scorer.score(sample), OPTIMAL_SCORE, atol=1e-2)  # type: ignore

    scorer = PropertyPredictorRegistry.get_property_predictor_scorer(
        property_name="charge",
        scorer_name="property_predictor_scorer",
        target=protein_further_ground_truths["charge"],
        parameters={"amide": "True", "ph": 5.0},
    )
    sample = select_sample("charge")
    assert isinstance(
        scorer, SCORING_FACTORY_WITH_PROPERTY_PREDICTORS["property_predictor_scorer"]
    )
    assert np.isclose(scorer.score(sample), OPTIMAL_SCORE, atol=1e-2)  # type: ignore
    assert len(PropertyPredictorRegistry.list_available_scorers()) == 3



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Cgcnn training utilities."""

import logging
import os
import shutil
import time
from dataclasses import dataclass, field
from random import sample
from typing import Any, Dict, Optional, Tuple, Union

import numpy as np
import torch
import torch.nn as nn
import torch.optim as optim
from sklearn import metrics
from torch import Tensor
from torch.autograd import Variable
from torch.optim.lr_scheduler import MultiStepLR
from torch.utils.data import DataLoader

from ...frameworks.cgcnn.data import CIFData, collate_pool, get_train_val_test_loader
from ...frameworks.cgcnn.model import CrystalGraphConvNet, Normalizer
from ..core import TrainingPipeline, TrainingPipelineArguments

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class CGCNNTrainingPipeline(TrainingPipeline):
    """CGCNN training pipelines for crystals."""

    def train(  # type: ignore
        self,
        training_args: Dict[str, Any],
        model_args: Dict[str, Any],
        dataset_args: Dict[str, Any],
    ) -> None:
        """Generic training function for CGCNN models.

        Args:
            training_args: training arguments passed to the configuration.
            model_args: model arguments passed to the configuration.
            dataset_args: dataset arguments passed to the configuration.

        Raises:
            NotImplementedError: the generic trainer does not implement the pipeline.
        """

        training_args["disable_cuda"] = (
            training_args["disable_cuda"] or not torch.cuda.is_available()
        )

        if training_args["task"] == "regression":
            best_mae_error = 1e10
        else:
            best_mae_error = 0.0

        # load data
        dataset = CIFData(dataset_args["datapath"])
        collate_fn = collate_pool
        train_loader, val_loader, test_loader = get_train_val_test_loader(  # type: ignore
            dataset=dataset,
            collate_fn=collate_fn,
            batch_size=training_args["batch_size"],
            num_workers=training_args["workers"],
            pin_memory=training_args["disable_cuda"],
            train_size=dataset_args["train_size"],
            val_size=dataset_args["val_size"],
            test_size=dataset_args["test_size"],
            return_test=True,
        )

        # obtain target value normalizer
        if training_args["task"] == "classification":
            normalizer = Normalizer(torch.zeros(2))
            normalizer.load_state_dict({"mean": 0.0, "std": 1.0})
        else:
            if len(dataset) < 500:
                logger.warning(
                    "Dataset has less than 500 data points. "
                    "Lower accuracy is expected. "
                )
                sample_data_list = [dataset[i] for i in range(len(dataset))]
            else:
                sample_data_list = [
                    dataset[i] for i in sample(range(len(dataset)), 500)
                ]
            _, sample_target, _ = collate_pool(sample_data_list)
            normalizer = Normalizer(sample_target)

        # build model
        structures, _, _ = dataset[0]
        orig_atom_fea_len = structures[0].shape[-1]
        nbr_fea_len = structures[1].shape[-1]  # type: ignore
        model = CrystalGraphConvNet(
            orig_atom_fea_len,
            nbr_fea_len,
            atom_fea_len=model_args["atom_fea_len"],
            n_conv=model_args["n_conv"],
            h_fea_len=model_args["h_fea_len"],
            n_h=model_args["n_h"],
            classification=True if training_args["task"] == "classification" else False,
        )
        if not training_args["disable_cuda"]:
            model.cuda()

        # define loss func and optimizer
        if training_args["task"] == "classification":
            criterion = nn.NLLLoss()
        else:
            criterion = nn.MSELoss()  # type: ignore

        if training_args["optim"] == "SGD":
            optimizer = optim.SGD(
                model.parameters(),
                training_args["lr"],
                momentum=training_args["momentum"],
                weight_decay=training_args["weight_decay"],
            )
        elif training_args["optim"] == "Adam":
            optimizer = optim.Adam(  # type: ignore
                model.parameters(),
                training_args["lr"],
                weight_decay=training_args["weight_decay"],
            )
        else:
            raise NameError("Only SGD or Adam is allowed as optimizer")

        # optionally resume from a checkpoint
        if training_args["resume"]:
            if os.path.isfile(training_args["resume"]):
                logger.info("loading checkpoint '{}'".format(training_args["resume"]))
                checkpoint = torch.load(training_args["resume"])
                training_args["start_epoch"] = checkpoint["epoch"]
                best_mae_error = checkpoint["best_mae_error"]
                model.load_state_dict(checkpoint["state_dict"])
                optimizer.load_state_dict(checkpoint["optimizer"])
                normalizer.load_state_dict(checkpoint["normalizer"])
                logger.info(
                    "loaded checkpoint '{}' (epoch {})".format(
                        training_args["resume"], checkpoint["epoch"]
                    )
                )
            else:
                logger.info(
                    "no checkpoint found at '{}'".format(training_args["resume"])
                )

        scheduler = MultiStepLR(
            optimizer, milestones=[training_args["lr_milestone"]], gamma=0.1
        )

        for epoch in range(training_args["start_epoch"], training_args["epochs"]):
            # train for one epoch
            train(
                train_loader,
                model,
                criterion,
                optimizer,
                epoch,
                normalizer,
                training_args["disable_cuda"],
                training_args["task"],
                training_args["print_freq"],
            )

            # evaluate on validation set
            mae_error = validate(
                val_loader,
                model,
                criterion,
                normalizer,
                training_args["disable_cuda"],
                training_args["task"],
                training_args["print_freq"],
                test=True,
            )

            if mae_error != mae_error:
                raise ValueError("mae_error is NaN")

            scheduler.step()

            # remember the best mae_eror and save checkpoint
            if training_args["task"] == "regression":
                is_best = mae_error < best_mae_error
                best_mae_error = min(mae_error, best_mae_error)
            else:
                is_best = mae_error > best_mae_error
                best_mae_error = max(mae_error, best_mae_error)
            save_checkpoint(
                {
                    "epoch": epoch + 1,
                    "state_dict": model.state_dict(),
                    "best_mae_error": best_mae_error,
                    "optimizer": optimizer.state_dict(),
                    "normalizer": normalizer.state_dict(),
                    "training_args": training_args,
                    "model_args": model_args,
                    "dataset_args": dataset_args,
                },
                is_best,
                training_args["output_path"],
            )

        # test best model
        logger.info("Evaluate Model on Test Set")
        best_path = os.path.join(training_args["output_path"], "model_best.pth.tar")
        if os.path.exists(best_path):
            best_checkpoint = torch.load(best_path)

            model.load_state_dict(best_checkpoint["state_dict"])
            validate(
                test_loader,
                model,
                criterion,
                normalizer,
                training_args["disable_cuda"],
                training_args["task"],
                training_args["print_freq"],
                test=True,
            )


def train(
    train_loader: Union[DataLoader[Any], Any],
    model: CrystalGraphConvNet,
    criterion: Union[nn.NLLLoss, nn.MSELoss],
    optimizer: Union[optim.SGD, optim.Adam],
    epoch: int,
    normalizer: Normalizer,
    disable_cuda: bool,
    task: str,
    print_freq: int,
) -> None:
    """Train step for cgcnn models.

    Args:
        train_loader: Dataloader for the training set.
        model: CGCNN model.
        criterion: loss function.
        optimizer: Optimizer to be used.
        epoch: Epoch number.
        normalizer: Normalize.
        disable_cuda: Disable CUDA.
        task: Training task.
        print_freq: Print frequency.
    """

    batch_time = AverageMeter()
    data_time = AverageMeter()
    losses = AverageMeter()
    mae_errors = AverageMeter()
    accuracies = AverageMeter()
    precisions = AverageMeter()
    recalls = AverageMeter()
    fscores = AverageMeter()
    auc_scores = AverageMeter()

    # switch to train mode
    model.train()

    end = time.time()
    for i, (input, target, _) in enumerate(train_loader):
        # measure data loading time
        data_time.update(time.time() - end)

        if not disable_cuda:
            input_var = (
                Variable(input[0].cuda(non_blocking=True)),
                Variable(input[1].cuda(non_blocking=True)),
                input[2].cuda(non_blocking=True),
                [crys_idx.cuda(non_blocking=True) for crys_idx in input[3]],
            )
        else:
            input_var = (Variable(input[0]), Variable(input[1]), input[2], input[3])
        # normalize target
        if task == "regression":
            target_normed = normalizer.norm(target)
        else:
            target_normed = target.view(-1).long()
        if not disable_cuda:
            target_var = Variable(target_normed.cuda(non_blocking=True))
        else:
            target_var = Variable(target_normed)

        # compute output
        output = model(*input_var)
        loss = criterion(output, target_var)

        # measure accuracy and record loss
        if task == "regression":
            mae_error = mae(normalizer.denorm(output.data.cpu()), target)
            losses.update(loss.data.cpu(), target.size(0))
            mae_errors.update(mae_error, target.size(0))  # type: ignore
        else:
            accuracy, precision, recall, fscore, auc_score = class_eval(
                output.data.cpu(), target
            )
            losses.update(loss.data.cpu().item(), target.size(0))
            accuracies.update(accuracy, target.size(0))
            precisions.update(precision, target.size(0))
            recalls.update(recall, target.size(0))
            fscores.update(fscore, target.size(0))
            auc_scores.update(auc_score, target.size(0))

        # compute gradient and do SGD step
        optimizer.zero_grad()
        loss.backward()
        optimizer.step()

        # measure elapsed time
        batch_time.update(time.time() - end)
        end = time.time()

        if i % print_freq == 0:
            if task == "regression":
                logger.info(
                    "Epoch: [{0}][{1}/{2}]\t"
                    "Time {batch_time.val:.3f} ({batch_time.avg:.3f})\t"
                    "Data {data_time.val:.3f} ({data_time.avg:.3f})\t"
                    "Loss {loss.val:.4f} ({loss.avg:.4f})\t"
                    "MAE {mae_errors.val:.3f} ({mae_errors.avg:.3f})".format(
                        epoch,
                        i,
                        len(train_loader),
                        batch_time=batch_time,
                        data_time=data_time,
                        loss=losses,
                        mae_errors=mae_errors,
                    )
                )
            else:
                logger.info(
                    "Epoch: [{0}][{1}/{2}]\t"
                    "Time {batch_time.val:.3f} ({batch_time.avg:.3f})\t"
                    "Data {data_time.val:.3f} ({data_time.avg:.3f})\t"
                    "Loss {loss.val:.4f} ({loss.avg:.4f})\t"
                    "Accu {accu.val:.3f} ({accu.avg:.3f})\t"
                    "Precision {prec.val:.3f} ({prec.avg:.3f})\t"
                    "Recall {recall.val:.3f} ({recall.avg:.3f})\t"
                    "F1 {f1.val:.3f} ({f1.avg:.3f})\t"
                    "AUC {auc.val:.3f} ({auc.avg:.3f})".format(
                        epoch,
                        i,
                        len(train_loader),
                        batch_time=batch_time,
                        data_time=data_time,
                        loss=losses,
                        accu=accuracies,
                        prec=precisions,
                        recall=recalls,
                        f1=fscores,
                        auc=auc_scores,
                    )
                )


def validate(
    val_loader: Union[DataLoader[Any], Any],
    model: CrystalGraphConvNet,
    criterion: Union[nn.MSELoss, nn.NLLLoss],
    normalizer: Normalizer,
    disable_cuda: bool,
    task: str,
    print_freq: int,
    test: bool = False,
) -> float:
    """Validation step for cgcnn models.

    Args:
        val_loader: Dataloader for the validation set.
        model: CGCNN model.
        criterion: loss function.
        normalizer: Normalize.
        disable_cuda: Disable CUDA.
        task: Training task.
        print_freq: Print frequency.
        test: test or only validate using the given dataset.

    Returns:
       average auc or mae depending on the training task.
    """

    batch_time = AverageMeter()
    losses = AverageMeter()
    mae_errors = AverageMeter()
    accuracies = AverageMeter()
    precisions = AverageMeter()
    recalls = AverageMeter()
    fscores = AverageMeter()
    auc_scores = AverageMeter()
    test_targets = []
    test_preds = []
    test_cif_ids = []

    # switch to evaluate mode
    model.eval()

    end = time.time()
    for i, (input, target, batch_cif_ids) in enumerate(val_loader):
        if not disable_cuda:
            with torch.no_grad():
                input_var = (
                    Variable(input[0].cuda(non_blocking=True)),
                    Variable(input[1].cuda(non_blocking=True)),
                    input[2].cuda(non_blocking=True),
                    [crys_idx.cuda(non_blocking=True) for crys_idx in input[3]],
                )
        else:
            with torch.no_grad():
                input_var = (Variable(input[0]), Variable(input[1]), input[2], input[3])
        if task == "regression":
            target_normed = normalizer.norm(target)
        else:
            target_normed = target.view(-1).long()
        if not disable_cuda:
            with torch.no_grad():
                target_var = Variable(target_normed.cuda(non_blocking=True))
        else:
            with torch.no_grad():
                target_var = Variable(target_normed)

        # compute output
        output = model(*input_var)
        loss = criterion(output, target_var)

        # measure accuracy and record loss
        if task == "regression":
            mae_error = mae(normalizer.denorm(output.data.cpu()), target)
            losses.update(loss.data.cpu().item(), target.size(0))
            mae_errors.update(mae_error, target.size(0))  # type: ignore
            if test:
                test_pred = normalizer.denorm(output.data.cpu())
                test_target = target
                test_preds += test_pred.view(-1).tolist()
                test_targets += test_target.view(-1).tolist()
                test_cif_ids += batch_cif_ids
        else:
            accuracy, precision, recall, fscore, auc_score = class_eval(
                output.data.cpu(), target
            )
            losses.update(loss.data.cpu().item(), target.size(0))
            accuracies.update(accuracy, target.size(0))
            precisions.update(precision, target.size(0))
            recalls.update(recall, target.size(0))
            fscores.update(fscore, target.size(0))
            auc_scores.update(auc_score, target.size(0))
            if test:
                test_pred = torch.exp(output.data.cpu())
                test_target = target
                assert test_pred.shape[1] == 2
                test_preds += test_pred[:, 1].tolist()
                test_targets += test_target.view(-1).tolist()
                test_cif_ids += batch_cif_ids

        # measure elapsed time
        batch_time.update(time.time() - end)
        end = time.time()

        if i % print_freq == 0:
            if task == "regression":
                logger.info(
                    "Test: [{0}/{1}]\t"
                    "Time {batch_time.val:.3f} ({batch_time.avg:.3f})\t"
                    "Loss {loss.val:.4f} ({loss.avg:.4f})\t"
                    "MAE {mae_errors.val:.3f} ({mae_errors.avg:.3f})".format(
                        i,
                        len(val_loader),
                        batch_time=batch_time,
                        loss=losses,
                        mae_errors=mae_errors,
                    )
                )
            else:
                logger.info(
                    "Test: [{0}/{1}]\t"
                    "Time {batch_time.val:.3f} ({batch_time.avg:.3f})\t"
                    "Loss {loss.val:.4f} ({loss.avg:.4f})\t"
                    "Accu {accu.val:.3f} ({accu.avg:.3f})\t"
                    "Precision {prec.val:.3f} ({prec.avg:.3f})\t"
                    "Recall {recall.val:.3f} ({recall.avg:.3f})\t"
                    "F1 {f1.val:.3f} ({f1.avg:.3f})\t"
                    "AUC {auc.val:.3f} ({auc.avg:.3f})".format(
                        i,
                        len(val_loader),
                        batch_time=batch_time,
                        loss=losses,
                        accu=accuracies,
                        prec=precisions,
                        recall=recalls,
                        f1=fscores,
                        auc=auc_scores,
                    )
                )

    if task == "regression":
        logger.info("MAE {mae_errors.avg:.3f}".format(mae_errors=mae_errors))
        return mae_errors.avg
    else:
        logger.info("AUC {auc.avg:.3f}".format(auc=auc_scores))
        return auc_scores.avg


def mae(prediction: Tensor, target: Tensor) -> Tensor:
    """Computes the mean absolute error between prediction and target.

    Args:
        prediction: torch.Tensor (N, 1)
        target: torch.Tensor (N, 1)

    Returns:
        the computed mean absolute error.
    """
    return torch.mean(torch.abs(target - prediction))


def class_eval(
    prediction: Tensor, target: Tensor
) -> Tuple[float, float, float, float, float]:
    """Class evaluation.

    Args:
        prediction: Predictions.
        target: Groundtruth.

    Returns:
        Computed accuracy, precision, recall, fscore, and auc_score.

    """

    prediction = np.exp(prediction.numpy())
    target = target.numpy()
    pred_label = np.argmax(prediction, axis=1)
    target_label = np.squeeze(target)
    if not target_label.shape:
        target_label = np.asarray([target_label])
    if prediction.shape[1] == 2:
        precision, recall, fscore, _ = metrics.precision_recall_fscore_support(
            target_label, pred_label, average="binary"
        )
        try:
            auc_score = metrics.roc_auc_score(target_label, prediction[:, 1])
        except ValueError:
            auc_score = 0.0
        accuracy = metrics.accuracy_score(target_label, pred_label)
    else:
        raise NotImplementedError
    return accuracy, precision, recall, fscore, auc_score


class AverageMeter:
    """Computes and stores the average and current value."""

    def __init__(self):
        """Initialize an AverageMeter object."""

        self.reset()

    def reset(self) -> None:
        """Reset values to 0."""

        self.val = 0.0
        self.avg = 0.0
        self.sum = 0.0
        self.count = 0

    def update(self, val: float, n: int = 1) -> None:
        """Update values of the AverageMeter.

        Args:
            val: value to be added.
            n: count.
        """
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count


def save_checkpoint(
    state: object, is_best: bool, path: str = ".", filename: str = "checkpoint.pth.tar"
) -> None:
    """Save CGCNN checkpoint.

    Args:
        state: checkpoint's object.
        is_best: whether the given checkpoint has the best performance or not.
        path: path to save the checkpoint.
        filename: checkpoint's filename.

    """

    torch.save(state, os.path.join(path, filename))
    if is_best:
        shutil.copyfile(
            os.path.join(path, filename), os.path.join(path, "model_best.pth.tar")
        )


@dataclass
class CGCNNDataArguments(TrainingPipelineArguments):
    """Data arguments related to CGCNN trainer."""

    __name__ = "dataset_args"

    datapath: str = field(
        metadata={
            "help": "Path to the dataset."
            "The dataset should follow the directory structure as described in https://github.com/txie-93/cgcnn"
        },
    )
    train_size: Optional[int] = field(
        default=None, metadata={"help": "Number of training data to be loaded."}
    )
    val_size: Optional[int] = field(
        default=None, metadata={"help": "Number of validation data to be loaded."}
    )
    test_size: Optional[int] = field(
        default=None, metadata={"help": "Number of testing data to be loaded."}
    )


@dataclass
class CGCNNModelArguments(TrainingPipelineArguments):
    """Model arguments related to CGCNN trainer."""

    __name__ = "model_args"

    atom_fea_len: int = field(
        default=64, metadata={"help": "Number of hidden atom features in conv layers."}
    )
    h_fea_len: int = field(
        default=128, metadata={"help": "Number of hidden features after pooling."}
    )
    n_conv: int = field(default=3, metadata={"help": "Number of conv layers."})
    n_h: int = field(
        default=1, metadata={"help": "Number of hidden layers after pooling."}
    )


@dataclass
class CGCNNTrainingArguments(TrainingPipelineArguments):
    """Training arguments related to CGCNN trainer."""

    __name__ = "training_args"

    task: str = field(
        default="regression",
        metadata={"help": "Select the type of the task."},
    )
    output_path: str = field(
        default=".",
        metadata={"help": "Path to the store the checkpoints."},
    )
    disable_cuda: bool = field(default=False, metadata={"help": "Disable CUDA."})
    workers: int = field(
        default=0, metadata={"help": "Number of data loading workers."}
    )
    epochs: int = field(default=30, metadata={"help": "Number of total epochs to run."})
    start_epoch: int = field(
        default=0, metadata={"help": "Manual epoch number (useful on restarts)."}
    )
    batch_size: int = field(default=256, metadata={"help": "Mini-batch size."})
    lr: float = field(default=0.01, metadata={"help": "Initial learning rate."})
    lr_milestone: float = field(
        default=100, metadata={"help": "Milestone for scheduler."}
    )
    momentum: float = field(default=0.9, metadata={"help": "Momentum."})
    weight_decay: float = field(default=0.0, metadata={"help": "Weight decay."})
    print_freq: int = field(default=10, metadata={"help": "Print frequency."})
    resume: str = field(default="", metadata={"help": "Path to latest checkpoint."})
    optim: str = field(default="SGD", metadata={"help": "Optimizer."})


@dataclass
class CGCNNSavingArguments(TrainingPipelineArguments):
    """Saving arguments related to CGCNN trainer."""

    __name__ = "saving_args"
